import ohio.ext.pandas
import pandas as pd
import numpy as np
import logging
import seaborn as sns
import matplotlib.table as tab
import matplotlib.pyplot as plt
import itertools
#from tabulate import tabulate

from IPython.display import display
from io import StringIO
from descriptors import cachedproperty
from sqlalchemy import create_engine
from sklearn.calibration import calibration_curve
from sklearn import metrics
from scipy.stats import spearmanr

from triage.component.catwalk.storage import ProjectStorage
from triage.component.postmodeling.error_analysis import generate_error_analysis, output_all_analysis
from triage.database_reflection import table_exists
from triage.component.catwalk.utils import sort_predictions_and_labels


class SingleModelAnalyzer:

    id_columns = ['entity_id', 'as_of_date']

    def __init__(self, model_id, engine):
        self.model_id=model_id
        self.engine=engine

    @cachedproperty
    def metadata(self):
        return next(self.engine.execute(
                    f'''
                    WITH individual_model_ids_metadata AS(
                        SELECT m.model_id,
                           m.model_group_id,
                           m.hyperparameters,
                           m.model_hash,
                           m.train_end_time::date,
                           m.train_matrix_uuid,
                           m.training_label_timespan,
                           m.model_type,
                           mg.model_config
                        FROM triage_metadata.models m
                        JOIN triage_metadata.model_groups mg
                        USING (model_group_id)
                        WHERE model_id = {self.model_id}
                    ),
                    individual_model_id_matrices AS(
                        SELECT DISTINCT ON (matrix_uuid)
                           model_id,
                           matrix_uuid,
                           evaluation_start_time as as_of_date
                        FROM test_results.evaluations
                        WHERE model_id = ANY(
                            SELECT model_id
                            FROM individual_model_ids_metadata
                        )
                    )
                    SELECT metadata.*, test.*
                    FROM individual_model_ids_metadata AS metadata
                    LEFT JOIN individual_model_id_matrices AS test
                    USING(model_id);'''
            )
        )
    
    @property
    def model_group_id(self):
        return self.metadata['model_group_id']

    @property
    def model_type(self):
        return self.metadata['model_type']

    @property
    def hyperparameters(self):
        return self.metadata['hyperparameters']

    @property
    def model_hash(self):
        return self.metadata['model_hash']

    @property
    def train_matrix_uuid(self):
        return self.metadata['train_matrix_uuid']

    # TODO: Need to figure out how this would work when there are multiple matrices in the evaluations table
    @property
    def pred_matrix_uuid(self):
        return self.metadata['matrix_uuid']

    @property
    def as_of_date(self):
        return self.metadata['as_of_date']

    @property
    def train_end_time(self):
        return self.metadata['train_end_time']

    @property
    def train_label_timespan(self):
        return self.metadata['training_label_timespan']

    def get_predictions(self, matrix_uuid=None, fetch_null_labels=True, predictions_table='test_results.predictions', subset_hash=None):
        """Fetch the predictions from the DB for a given matrix
        
            args:
                matrix_uuid (optional):  
        
        """
        where_clause = f"WHERE model_id = {self.model_id}"

        if matrix_uuid is not None:
            where_clause += f" AND matrix_uuid='{matrix_uuid}'"

        if not fetch_null_labels:
            where_clause += f" AND label_value IS NOT NULL"

        if subset_hash:
            # get subset table name
            q = f"select config from triage_metadata.subsets where subset_hash='{subset_hash}'"
            config_df = pd.read_sql(q, self.engine)
            table_name = f"subset_{config_df.iloc[0]['config']['name']}_{subset_hash}"
            predictions_table += f" preds join {table_name} subset on preds.entity_id = subset.entity_id and preds.as_of_date = subset.as_of_date"

        query = f"""
            SELECT model_id,
                   entity_id,
                   as_of_date,
                   score,
                   label_value,
                   rank_abs_with_ties,
                   rank_pct_with_ties,
                   rank_abs_no_ties,
                   rank_pct_no_ties, 
                   test_label_timespan
            FROM {predictions_table}
            {where_clause}        
        """

        preds = pd.read_sql(query, self.engine)

        #TODO: Maybe we should call the script to save predictions here?
        if preds.empty:
            logging.warning(f'No predictions were found in {predictions_table} for model_id {self.model_id}. Returning empty dataframe!')
            # raise RuntimeError(
                # "No predictions were found in the database. Please run the add_predictions module to add predictions for the model"
            # )
            return preds 
        

        return preds.set_index(self.id_columns)

    
    def get_top_k(self, threshold_type, threshold, matrix_uuid=None):
        """ Fetch the k intities with highest model score for a given model
        
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                    - rank_pct_no_ties 
                    - rank_pct_with_ties
                    - rank_abs_no_ties
                    - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                matrix_uuid (optional, str): If a list should be generated out of a matrix that were not used to validate the model during the experiment
        """

        q = f"""
            select 
                entity_id, 
                as_of_date,
                score,
                label_value,
                rank_abs_no_ties,
                rank_abs_with_ties,
                rank_pct_no_ties,
                rank_pct_with_ties,
                matrix_uuid
            from test_results.predictions
            where model_id={self.model_id}
            and {threshold_type} <= {threshold}
        """

        top_k = pd.read_sql(q, self.engine)

        return top_k


    # TODO generalize this function
    # TODO do we want to allow specifying a dictionary of attributes + values? 
    def get_aequitas(self, parameter=None, attribute_name=None, subset_hash=None):
        '''
        Get aequitas evaluations from the DB

        Args:
            parameter (str): Optional. The threshold to apply when returning the aequitas evaluations.
                                  If not specified, all aequitas evaluations will be returned.

            attribute_name (str): Optional. Fetch aequitas evaluations related to a particular attribute.

            subset_hash (str): Optional. For fetching evaluations of a specific subset.    
        '''

        where_clause = f'WHERE model_id={self.model_id}'

        if subset_hash is not None:
            where_clause += f" AND 'subset_hash='{subset_hash}'"
        else:
            where_clause += f" AND subset_hash=''"

        if parameter is not None:
            where_clause += f" AND parameter='{parameter}'"

        if attribute_name:
            where_clause += f" AND attribute_name='{attribute_name}'"
        
        # TODO don't return all columns ?
        q = f"""
            select
                * 
            from test_results.aequitas
            {where_clause}
        """

        evaluations = pd.read_sql(q, self.engine)

        return evaluations
    

    def get_evaluations(self, metrics=None, matrix_uuid=None, subset_hash=None):
        ''' 
        Get evaluations for the model from the DB

        Args:
            metrics Dict[str:List]): Optional. The metrics and parameters for evaluations. 
                                    A dictionary of type {metric:[thresholds]}
                                    If not specified, all the evaluations will be returned

            matrix_uuid (str): Optional. If model was evaluated using multiple matrices
                            one could get evaluations of a specific matrix. Defaults to fetching everything

            subset_hash (str): Optional. For fetching evaluations of a specific subset.    
        '''

        where_clause = f'WHERE model_id={self.model_id}'

        if matrix_uuid is not None:
            where_clause += f" AND matrix_uuid='{matrix_uuid}'"

        if subset_hash is not None:
            where_clause += f" AND subset_hash='{subset_hash}'"
        else:
            where_clause += f" AND subset_hash=''"

        if metrics is not None:
            where_clause += " AND ("
            for i, metric in enumerate(metrics):
                parameters = metrics[metric]
                where_clause += f""" metric='{metric}' AND parameter in ('{"','".join(parameters)}')"""

                if i < len(metrics) - 1:
                    where_clause += "OR"

            where_clause += ") "

        q = f"""
            select
                model_id,
                matrix_uuid,
                subset_hash,
                metric, 
                parameter,
                stochastic_value,
                num_labeled_above_threshold,               
                num_positive_labels
            from test_results.evaluations
            {where_clause}
        """

        evaluations = pd.read_sql(q, self.engine)

        return evaluations

    def get_feature_importances(self, n_top_features=20):

        logging.debug(f'Fetching feature importance from db for model id: {self.model_id}')
        features = pd.read_sql(
           f'''
           select
                feature,
                feature_importance,
                rank_abs
           FROM train_results.feature_importances
           WHERE model_id = {self.model_id}
           and rank_abs <= {n_top_features}
           and abs(feature_importance) > 0 
           order by rank_abs
           ''', con=self.engine)
        return features
    
    def get_feature_group_importances(self):
        """
        Get the top most important feature groups as identified by the maximum importance of any feature in the group

        """
        # TODO this assumes any experiment linked to this model has the same feature aggregations, is this valid?
        q = f"""
            select distinct experiment_hash from triage_metadata.models m 
            left join triage_metadata.experiment_models em on m.model_hash=em.model_hash 
            where model_group_id={self.model_group_id}
            """
        experiment_hashes = pd.read_sql(q, self.engine)
        experiment_hash = experiment_hashes['experiment_hash'].iloc[0]

        # get feature group names
        q = f"""
            select 
                config->'feature_aggregations' as feature_groups
            from triage_metadata.experiments where experiment_hash = '{experiment_hash}'
        """

        feature_groups = [i['prefix'] for i in pd.read_sql(q, self.engine)['feature_groups'].iloc[0]]
        feature_groups
        case_part = ''
        for fg in feature_groups:
            case_part = case_part + "\nWHEN feature like '{fg}%%' THEN '{fg}'".format(fg=fg)

        # get feature group importances
        feature_group_importance = pd.read_sql(f"""
            with raw_importances as (
                select 
                    model_id,
                    feature,
                    feature_importance,
                    CASE {case_part}
                    ELSE 'No feature group'
                    END as feature_group
                FROM train_results.feature_importances
                WHERE model_id = {self.model_id}
            )
            SELECT
            model_id,
            feature_group,
                max(abs(feature_importance)) as importance_aggregate
            FROM raw_importances
            GROUP BY feature_group, model_id
        """, con=self.engine)
        return feature_group_importance

    def crosstabs_pos_vs_neg(self, project_path, thresholds, matrix_uuid=None, push_to_db=True, table_name='crosstabs', return_df=True, replace=True, predictions_table='test_results.predictions'):
        """ Generate crosstabs for the predicted positives (top-k) vs the rest
    
        args:
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
            return_df (bool, optional): Whether to return the constructed df or just to store in the database
                                        Defaults to False (only writing to the db)
            
            matrix_uuid (str, optional): To run crosstabs for a different matrix than the validation matrix from the experiment

            push_to_db (bool, optional): Whether to write the results to the database. Defaults to True
            table_name (str, optional): Table name to use in the db's `test_results` schema. Defaults to crosstabs.
                                        If the table exists, results are appended

            return_df (bool, optional): Whether to return the crosstabs as a dataframe. Defaults to True
        """

        # metrics we calculate for positive and negative predictions
        positive_mean = lambda pos, neg: pos.mean(axis=0)
        negative_mean = lambda pos, neg: neg.mean(axis=0)
        positive_std = lambda pos, neg: pos.std(axis=0)
        negative_std = lambda pos, neg: neg.std(axis=0)
        ratio_positive_negative = lambda pos, neg: pos.mean(axis=0) / neg.mean(axis=0)
        # mean_ratio = lambda pos, neg: max(ratio_positive_negative(pos, neg).at[0], (float(1) / ratio_positive_negative(pos, neg)).at[0])
        positive_support = lambda pos, neg: (pos > 0).sum(axis=0)
        negative_support = lambda pos, neg: (neg > 0).sum(axis=0)
        positive_support_pct = lambda pos, neg: round((pos > 0).sum(axis=0).astype(float) / len(pos), 3)
        negative_support_pct = lambda pos, neg: round((neg > 0).sum(axis=0).astype(float) / len(neg), 3)

        def mean_ratio(pos, neg):
            pos_over_neg = ratio_positive_negative(pos, neg)
            neg_over_pos = float(1) / pos_over_neg

            df = pd.DataFrame([pos_over_neg, neg_over_pos])

            return df.max()

        crosstab_functions = [
            ("mean_predicted_positive", positive_mean),
            ("mean_predicted_negative", negative_mean),
            ("std_predicted_positive", positive_std),
            ("std_predicted_negative", negative_std),
            ("mean_ratio", mean_ratio),
            ("support_predicted_positive", positive_support),
            ("support_predicted_negative", negative_support),
            ("support_pct_predicted_positive", positive_support_pct),
            ("support_pct_predicted_negative", negative_support_pct)
        ]


        if matrix_uuid is None:
            matrix_uuid = self.pred_matrix_uuid
            logging.debug(f'Matrix uuid set to: {matrix_uuid}')

        predictions = self.get_predictions(matrix_uuid=matrix_uuid, predictions_table=predictions_table)

    
        if predictions.empty:
            logging.error(f'No predictions found for {self.model_id} and {matrix_uuid}. Exiting!')
            raise ValueError(f'No predictions found {self.model_id} and {matrix_uuid}')

        # Check whether the table exists
        if table_exists(f'test_results.{table_name}', self.engine):
            # checking whether the crosstabs already exist for the model
            logging.debug(f'Checking whether crosstabs already exist for the model {self.model_id} and {matrix_uuid}')
            q = f"select * from test_results.{table_name} where model_id={self.model_id} and matrix_uuid='{matrix_uuid}';"
            df = pd.read_sql(q, self.engine)
            
            if not df.empty:
                logging.warning(f'Crosstabs aleady exist for model {self.model_id} and matrix_uuid={matrix_uuid}')

                if replace:
                    logging.warning('Deleting the existing crosstabs!')
                    with self.engine.connect() as conn:
                        conn.execute(f"delete from test_results.{table_name} where model_id={self.model_id} and matrix_uuid='{matrix_uuid}';")
                else:
                    logging.info(f"Replace set to False. Not calculating crosstabs for model {self.model_id} and matrix_uuid='{matrix_uuid}';")
                    if return_df: return df 
                    else: return 
            
        # initializing the storage engines
        project_storage = ProjectStorage(project_path)
        matrix_storage_engine = project_storage.matrix_storage_engine()

        matrix_store = matrix_storage_engine.get_store(matrix_uuid=matrix_uuid)

        matrix = matrix_store.design_matrix
        
        labels = matrix_store.labels
        features = matrix.columns

        # joining the predictions to the model
        matrix = predictions.join(matrix, how='left')

        all_results = list() 
        
        for threshold_name, threshold in thresholds.items():
            logging.info(f'Crosstabs using threshold: {threshold_name} <= {threshold}')

            msk = matrix[threshold_name] <= threshold
            postive_preds = matrix[msk][features]
            negative_preds = matrix[~msk][features]

            temp_results = list()
            for name, func in crosstab_functions:
                logging.info(name)

                this_result = pd.DataFrame(func(postive_preds, negative_preds))
                this_result['metric'] = name
                temp_results.append(this_result)
            
            temp_results = pd.concat(temp_results)
            temp_results['threshold_type'] = threshold_name
            temp_results['threshold'] = threshold

            all_results.append(temp_results)

        crosstabs_df = pd.concat(all_results).reset_index()

        crosstabs_df.rename(columns={'index': 'feature', 0: 'value'}, inplace=True)
        crosstabs_df['model_id'] = self.model_id
        crosstabs_df['matrix_uuid'] = matrix_uuid

        if push_to_db:
            # logging.info('Pushing the results to the DB')
            # crosstabs_df.set_index(
            #     ['model_id', 'matrix_uuid', 'feature', 'metric', 'threshold_type', 'threshold'], inplace=True
            # )

            # # TODO: Figure out to change the owner of the table
            # crosstabs_df.pg_copy_to(schema='test_results', name=table_name, con=self.engine, if_exists='append')
            logging.info(f'Pushing the results to the database, {len(crosstabs_df)} rows')
                    
            crosstabs_df.set_index(
                ['model_id', 'matrix_uuid', 'feature', 'metric', 'threshold_type', 'threshold'],
                inplace=True
            )
            
            crosstabs_df = crosstabs_df.reset_index()
            
            if not table_exists(f'test_results.{table_name}', self.engine):
                q = f'''
                    create schema if not exists test_results;
                    
                    create table test_results.{table_name} (
                    model_id INTEGER,
                    matrix_uuid TEXT,
                    feature TEXT,
                    metric TEXT,
                    threshold_type TEXT,
                    threshold FLOAT,
                    value FLOAT  
                    );
                
                '''
                # q = _generate_create_table_sql_statement_from_df(results, f'{table_schema}.{table_name}')
                self.engine.execute(q)
            
            conn = self.engine.raw_connection()
            cursor = conn.cursor()
            
            buffer = StringIO()
            crosstabs_df.to_csv(buffer, index=False, header=False)
            buffer.seek(0)
            
            columns = ', '.join(crosstabs_df.columns)
            print(columns)
            cursor.copy_expert(f"COPY test_results.{table_name} ({columns}) FROM STDIN WITH CSV", buffer)
            # results.to_sql(con=db_engine, schema=table_schema, name=table_name, if_exists='append')
            conn.commit()
            cursor.close()
            conn.close()

        if return_df:
            return crosstabs_df
        

    def display_crosstabs_pos_vs_neg(
            self, 
            threshold_type, threshold, table_name='crosstabs',
            display_n_features=40,
            filter_features=None,
            support_threshold=0.1,
            return_df=True,
            show_plot=True,
            matrix_uuid=None,
            ax=None):
        """ Display the crosstabs for the model

        Args:
            threshold_type (str): Type of rank threshold to use in splitting predicted positives from negatives. 
                                Has to be one of the rank columns in the test_results.predictions_table

            threshold (Union[int, float]): The rank threshold of the specified type. If the threshold type is an absolute, integer. 
                                        If percentage, should be a float between 0 and 1
        
            table_name (str, optional): Table name to fetch crosstabs from (`test_results` schema). Defaults to 'crosstabs'.
        """

        q = f"""
            select 
                model_id, 
                feature, 
                metric, 
                abs(value) as value 
            from test_results.{table_name} 
            where threshold_type = '{threshold_type}'
            and threshold = {threshold}
            and metric in (
                'mean_ratio',
                'mean_predicted_positive',
                'mean_predicted_negative',
                'support_pct_predicted_positive',
                'support_pct_predicted_negative'
            )
            and model_id = {self.model_id}            
        """

        if matrix_uuid is not None:
            q = q + f"\n and matrix_uuid='{matrix_uuid}'"
        else:
            matrix_uuid=self.pred_matrix_uuid

        ct = pd.read_sql(q, self.engine)

        if ct.empty:
            logging.error(
                f'''Crosstabs not found for model {self.model_id} and matrix={matrix_uuid} in table test_results.{table_name}. 
                Please use crosstabs_pos_vs_neg function to generate crosstabs'''
            )
            raise ValueError('Crosstabs not found!')

        # Creating the pivot table to convert to wide format indexed by the column
        pivot_table = pd.pivot(
            ct, 
            index='feature', 
            columns='metric', 
            values='value'
        )

        # Shortening the names, and removing the index names to plot more cleanly
        pivot_table.rename(
            columns={
                'mean_ratio': 'ratio',
                'mean_predicted_positive': '(+)mean',
                'mean_predicted_negative': '(-)mean',
                'support_pct_predicted_negative': '(-)supp',
                'support_pct_predicted_positive': '(+)supp'
            }, 
            inplace=True
        )
        pivot_table.index.name='feature_name'
        pivot_table.columns.name=''

        if filter_features is not None:
            return pivot_table.loc[filter_features]

        
        # Filtering by the support threshold
        msk1 = pivot_table['(+)supp'] > support_threshold
        # Features with highest postive : negative ratio
        # df1 = pivot_table[msk].sort_values(
        #     'ratio', 
        #      ascending = False
        # ).head(display_n_features)


        msk2 = pivot_table['(-)supp'] > support_threshold
        # Features with the highest negative : positive ratio
        # df2 = pivot_table[msk].sort_values(
        #     ['ratio', '(-)supp'], 
        #      ascending = [True, False]
        # ).head(display_n_features)

        df = pivot_table[msk1 | msk2].sort_values(
            'ratio', 
             ascending = False
        ).head(display_n_features)

        if show_plot:
            if ax is None: 
                fig, ax = plt.subplots(figsize=(4, 1 + (display_n_features *2) / 5), dpi=100)

            t = tab.table(
                ax, 
                cellText=df.values.round(1), 
                colLabels=df.columns, 
                rowLoc='right',
                rowLabels=df.index, 
                bbox=[0, 0, 1, 1]
            )
        
            ax.set_title(
                f'Model Group: {self.model_group_id} | Train end: {self.train_end_time} | Model: {self.model_id}\n',
                fontdict={
                    'fontsize': 10,
                    'verticalalignment': 'center',
                    'horizontalalignment': 'center'
                },
                x=-0.1
            )
            ax.axis('off')

            # Table formatting
            t.auto_set_column_width([0, 1, 2, 3, 4])
            for key, cell in t.get_celld().items():
                cell.set_fontsize(9)
                cell.set_linewidth(0.1)
                # cell.PAD = 0.01
                
        
        title_str = f'model_group: {self.model_group_id} | train_end_time: {self.train_end_time} | model_id: {self.model_id}'
        # print(title_str)
        # print(tabulate(df.round(1), headers='keys', tablefmt='RST'))


        if return_df:
            df.style.set_caption(title_str)
            return df    
        

    def error_analysis(self, project_path):
        """
        Generates three main error anlaysis with Decision Trees in order to identity 
        what are the features with most importance when the model make mistakes 
        in negative labels (FN), positive labels (FP) and both, negative and 
        positive label errors (FN & FP).

        Args:
            project_path (string): Path for the output of the project 

        Returns: 
            error_analysis_results (list): List of dictionaries with all the error
            analysis made.
        """
        model_id = self.model_id
        connection = self.engine
        error_analysis_results = generate_error_analysis(model_id, connection, project_path=project_path)
        #TODO do we want to 
        return error_analysis_results


    def plot_score_distribution(self, ax, nbins=10, top_k=None, matrix_uuid=None):
        """
        Plot the distribution of predicted scores for all entities across model groups and train_end_time
        Optionally, show only the top k predicted scores.

        Args:
            ax: matplotlib Axes object to plot on
            nbins (optional, int): the number of bins to apply to the score histogram
            top_k (optional, int): if not None, displays only the top_k scores.
            matrix_uuid (optional): get model scores for a particular matrix
        """

        predictions = self.get_predictions(matrix_uuid)
        # keep only top_k if specified
        if top_k:
            predictions = predictions.loc[predictions['rank_abs_no_ties'] <= top_k]
        ax.hist(predictions.score,
                 bins=nbins,
                 alpha=0.5,
                 color='blue')
        ax.axvline(predictions.score.mean(),
                    color='black',
                    linestyle='dashed')
        ax.set_ylabel('Frequency')
        ax.set_xlabel('Score')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')

        return ax

    def plot_score_label_distribution(self, ax,
                                      nbins=10,
                                      top_k=None,
                                      matrix_uuid=None, 
                                      label_names = ('Label = 0', 'Label = 1')):
        """
        Plot the distribution of predicted scores for all entities across model groups and train_end_time
        Optionally, show only the top k predicted scores.
        NOTE: this function only handles 0/1 labels, ignores null labels

        Args:
            ax: matplotlib Axes object to plot on
            nbins (int): the number of bins to define the score distribution
            top_k (optional, int): if not None, displays only the top_k scores.
            matrix_uuid (optional): specify a matrix to get predictions for
            label_names (tuple[String]): specify the two label names to display on the plot

        Return:
            ax: the modified Axes object
        """

        df_predictions = self.get_predictions(matrix_uuid)
        # keep only top_k if specified
        if top_k:
            df_predictions = df_predictions.loc[df_predictions['rank_abs_no_ties'] <= top_k]
        df__0 = df_predictions[df_predictions.label_value == 0]
        df__1 = df_predictions[df_predictions.label_value == 1]

        ax.hist(df__0.score,
                 bins=nbins,
                 alpha=0.5,
                 color='skyblue',
                 label=label_names[0])
        ax.hist(list(df__1.score),
                 bins=nbins,
                 alpha=0.5,
                 color='orange',
                 label=label_names[1])
        ax.axvline(df__0.score.mean(),
                    color='skyblue',
                    linestyle='dashed')
        ax.axvline(df__1.score.mean(),
                    color='orange',
                    linestyle='dashed')
        ax.legend(bbox_to_anchor=(0., 1.005, 1., .102),
                   loc=8,
                   ncol=2,
                   borderaxespad=0.)
        ax.set_ylabel('Frequency')
        ax.set_xlabel('Score')
        ax.set_title('Score Distribution across Labels')
        return ax
    
    # TODO make this more general purpose
    def plot_bias_threshold_curve(self, attribute_name, attribute_value, bias_metric, ax):
        bias_df = self.get_aequitas()
        # limit to specific attribute+value
        bias_df = bias_df.loc[(bias_df['attribute_name']==attribute_name) & (bias_df['attribute_value']==attribute_value)]

        bias_df['perc_points'] = [x.split('_')[0] for x in bias_df['parameter'].tolist()]
        bias_df['perc_points'] = pd.to_numeric(bias_df['perc_points'])

        msk_pct = bias_df['parameter'].str.contains('pct')

        # plot precision
        sns.lineplot(
            x='perc_points',
            y=bias_metric, 
            data=bias_df[msk_pct], 
            label=self.model_group_id,
            ax=ax, 
            estimator='mean', ci='sd'
        )
        ax.set_xlabel('List size percentage (k%)')
        #ax.set_ylabel(f'{bias_metric} for {attribute_name}:{attribute_value}')
        return ax

    def plot_precision_threshold_curve(self, ax, matrix_uuid=None):
        eval_df = self.get_evaluations(matrix_uuid=matrix_uuid)

        eval_df['perc_points'] = [x.split('_')[0] for x in eval_df['parameter'].tolist()]
        eval_df['perc_points'] = pd.to_numeric(eval_df['perc_points'])

        msk_prec = eval_df['metric']=='precision@'
        msk_pct = eval_df['parameter'].str.contains('pct')

        # plot precision
        sns.lineplot(
            x='perc_points',
            y='stochastic_value', 
            data=eval_df[msk_pct & msk_prec], 
            label=self.model_group_id,
            ax=ax, 
            estimator='mean', ci='sd'
        )
        ax.set_xlabel('List size percentage (k%)')
        ax.set_ylabel('Precision')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        return ax

    # def plot_recall_curve(self, ax=None, matrix_uuid=None, list_size_upper_bound_pct=1, pct_step_size=0.01, subset_hash=None, title_string=None):
    #     pred_df = self.get_predictions(matrix_uuid=matrix_uuid)
        
        
    # TODO: Facilitate plotting pr-k with absolute thresholds
    def plot_precision_recall_curve(self, ax=None, matrix_uuid=None, list_size_upper_bound_pct=1, pct_step_size=0.01, subset_hash=None, title_string=None, only_recall=False):
        """
        Plots precision-recall curves at each train_end_time for all model groups
        
        Args:
            ax: matplotlib Axes object to plot on
            matrix_uuid (optional): specify a matrix to get predictions for
            
        Return:
            ax: the modified Axes object
        """

        pred_df = self.get_predictions(matrix_uuid=matrix_uuid)
        # remove null labels
        pred_df.dropna(axis=0, subset=['label_value'], inplace=True) 
        # handle ties
        pred_df_score, pred_df_label, pred_df_index = sort_predictions_and_labels(pred_df['score'], pred_df['label_value'], pred_df.index, tiebreaker="random", sort_seed=15321)
        pred_df = pd.DataFrame(list(zip(pred_df_score, pred_df_label)), columns=['score', 'label_value'])
        
        if ax is None:
            fig, ax = plt.subplots()


        if pred_df.empty or subset_hash is not None:
            logging.warning(f'''
                            No predictions were found for model id {self.model_id} (group: {self.model_group_id}) 
                            or a subset_hash was provided. Using the evaluations to generate the plot. Zoomed PR-K not supported!'''
                            )
            eval_df = self.get_evaluations(matrix_uuid=matrix_uuid, subset_hash=subset_hash)
            
            if eval_df.empty: 
                logging.error('No evaluations were found! Returning empty axes!')
                return ax
            
            eval_df['perc_points'] = [x.split('_')[0] for x in eval_df['parameter'].tolist()]
            eval_df['perc_points'] = pd.to_numeric(eval_df['perc_points'])
            
            msk_prec = eval_df['metric']=='precision@'
            msk_recall = eval_df['metric']=='recall@'
            msk_pct = eval_df['parameter'].str.contains('pct')

            # plot precision
            if not only_recall:
                sns.lineplot(
                    x='perc_points',
                    y='stochastic_value', 
                    data=eval_df[msk_pct & msk_prec], 
                    label='precision@k',
                    ax=ax, 
                    estimator='mean', ci='sd'
                )
            # plot recall
            sns.lineplot(
                x='perc_points', 
                y='stochastic_value', 
                data=eval_df[msk_pct & msk_recall], 
                label='recall@k', 
                ax=ax, 
                estimator='mean', 
                ci='sd'
            )

        else:
            logging.debug(f'Found saved predictions for model id {self.model_id}.(group: {self.model_group_id})')
            k_values = np.arange(0, list_size_upper_bound_pct + pct_step_size, pct_step_size)

            precisions = list()
            recalls = list()
            num_scored = len(pred_df)
            num_positives = pred_df.label_value.sum()

            for k in k_values:
                num_above_thresh = round(k * num_scored)

                pred_pos = pred_df.iloc[:num_above_thresh]
                
                
                precision = pred_pos.label_value.sum() / num_above_thresh if num_above_thresh > 0 else 0
                recall = pred_pos.label_value.sum() / num_positives if num_above_thresh > 0 else 0

                precisions.append(precision)
                recalls.append(recall)
                
            with sns.axes_style("darkgrid"): 
                if not only_recall:
                    sns.lineplot(x=k_values * 100, y=precisions, ax=ax, label='Precision@k')
                sns.lineplot(x=k_values * 100, y=recalls, ax=ax, label='Recall@k')
            

        ax.set_xlabel('Population percentage (k %)')
        ax.set_ylabel('Metric Value')
        ax.set_xlim(0, 100)
        ax.set_ylim(0, 1)
        # ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        # ax.set_title('Precision-Recall Curve')
        
        ax.legend(frameon=False)
        
        if title_string is None:
            ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        else:
            ax.set_title(title_string)
        return ax

    def plot_feature_importance(self, ax, n_top_features=20):
        """
        Plot the top most important individual features across model groups and train end times

        Args:
            ax: matplotlib Axes object to plot on
            n_top_features (int): the number of features to display
            
        Return:
            ax: the modified Axes object
        """

        feature_importance_scores = self.get_feature_importances(n_top_features=n_top_features)
        # keep only top n_top_features
        # feature_importance_scores.sort_values(by=['rank_abs'], ascending=True, inplace=True)
        # feature_importance_scores = feature_importance_scores.loc[feature_importance_scores['rank_abs'] <= n_top_features]
        
        if feature_importance_scores.empty:
            return ax


        if 'Algorithm does not support' in feature_importance_scores.feature.iloc[0]:
            # For models without featue importance score support
            return ax 
        
        # plot
        sns.barplot(
            data=feature_importance_scores,
            x='feature_importance', 
            y='feature',
            color='royalblue',
            ax=ax
        )
        ax.set_ylabel('Feature')
        ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        return ax

    def plot_feature_group_importance(self, ax, n_top_groups=20):
        """
        Plot the top most important feature groups as identified by the maximum importance of any feature in the group

        Args:
            ax: matplotlib Axes object to plot on

        Return: 
            ax: modified matplotlib Axes object
        """
        feature_group_importance = self.get_feature_group_importances()
        feature_group_importance.sort_values(by=['importance_aggregate'], ascending=False, inplace=True)
        feature_group_importance = feature_group_importance.reset_index(drop=True)
        feature_group_importance = feature_group_importance[feature_group_importance.index < n_top_groups] 
        sns.barplot(
            data=feature_group_importance,
            x='importance_aggregate', 
            y='feature_group',
            color='royalblue',
            ax=ax
        )
        ax.set_ylabel('Feature Group')
        ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        return ax

    def plot_calibration_curve(self, ax, nbins=20, min_score=0, max_score=1, matrix_uuid=None):

        """
        Plot the calibration curve of predicted probability scores across model groups

        Args:
            ax: matplotlib Axes object to plot on
            nbins (int): the number of bins to define the calibration curve
            min_score (optional, float between 0 and 1): zoom in on the plot, set x and y min axis value
            max_score (optional, float between 0 and 1): zoom in on the plot, set x and y max axis value
            matrix_uuid (optional): specify a matrix to get predictions for

        Return: 
            ax: modified matplotlib Axes object
        """
        # TODO can the bins be adjusted when zooming in on the plot?
        scores = self.get_predictions(matrix_uuid) # TODO what about null labels?
        scores = scores.dropna()
        # TODO do we want to restrict to label=1 only?
        cal_x, cal_y = calibration_curve(scores['label_value'], scores['score'], n_bins=nbins)
        sns.lineplot(
            x=cal_x,
            y=cal_y, 
            marker='o', 
            ax=ax
        )
        # plot perfectly calibrated line y = x for comparison
        probabilities = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        # filter out probabilities
        probabilities_plot = []
        for p in probabilities:
            probabilities_plot.append(p)
        sns.lineplot(
            data = pd.DataFrame(list(zip(probabilities, probabilities)), columns=['x', 'y']),
            x = 'x', 
            y = 'y', 
            color = 'gray',
            linestyle='dashed',
            ax=ax
        )
        ax.set_xlabel('Mean predicted probability')
        ax.set_ylabel('Fraction of data')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        ax.set_xlim([min_score, max_score])
        ax.set_ylim([min_score, max_score])
        return ax
    

        
"""This class is still WIP. Proceed with caution"""
class MultiModelAnalyzer: 

    def __init__(self, engine, model_groups, experiment_hashes, project_path=None, train_end_times=None, use_all_model_groups=False) -> None:
        self.model_groups = model_groups
        self.experiment_hashes = experiment_hashes # TODO made experiment hashes into a list to plot models from different experiments for MVESC, there's probably a better way to generalize this
        self.engine = engine
        self.project_path = project_path
        if use_all_model_groups: # shortcut to listing out all model groups for an experiment
            self.use_all_model_groups()
        self.models = self.get_model_ids()

        #TODO: Would be good to let the user restrict the train_end_times included in the report

        
    @property
    def model_ids(self):
        model_ids = list()
        for mg, d in self.models.items():
            for t, mod_obj in d.items():
                model_ids.append(mod_obj.model_id)

    @property
    def model_types(self):
        pass

    def use_all_model_groups(self):
        experiment_hashes = "','".join(self.experiment_hashes)
        q = f"""
                select model_group_id 
                from triage_metadata.models
                    join triage_metadata.experiment_models using (model_hash)
                where experiment_hash in ('{experiment_hashes}')
            """
        model_groups = pd.read_sql(q, self.engine)
        self.model_groups = model_groups['model_group_id'].to_list()

    # TODO: revise this to show the only the model_group_id, list of model ids, model type, and hyperparameters
    def display_model_groups(self):
        data_dict = []
        for mg in self.model_groups:
            for train_end_time in self.models[mg]:
                model_analyzer = self.models[mg][train_end_time]
                data_dict.append([mg, train_end_time, model_analyzer.model_id, model_analyzer.model_type, model_analyzer.hyperparameters])
        all_models = pd.DataFrame(data_dict, columns=['model_group_id', 'train_end_time', 'model_id', 'model_type', 'hyperparameters'])

        # displaying the model_group_ids, model_type, and the hyperparameters
        to_print = all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().to_dict(orient='records')

        for m in to_print:
            print(m)

        return all_models
    
    def print_model_summary(self):
        ''' This is mostly to be used as a key for modeling report plots (just as a model group number to model group mapping)'''
        data_dict = []
        for mg in self.model_groups:
            for train_end_time in self.models[mg]:
                model_analyzer = self.models[mg][train_end_time]
                data_dict.append([mg, train_end_time, model_analyzer.model_id, model_analyzer.model_type, model_analyzer.hyperparameters])
        
        all_models = pd.DataFrame(data_dict, columns=['model_group_id', 'train_end_time', 'model_id', 'model_type', 'hyperparameters'])
        # to_print = all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().to_dict(orient='records')

        for i, model in all_models.groupby('model_group_id').nth(1)[['model_type', 'hyperparameters']].reset_index().iterrows():
            print(f"{model['model_group_id']} - {model['model_type']} with ({model['hyperparameters']}) ")

    def cohort_summary(self):
        q = f"""
            select distinct on(train_end_time)
                -- matrix_uuid,
                evaluation_start_time as train_end_time,
                num_labeled_examples as cohort_size,
                num_positive_labels,
                case when num_labeled_examples > 0 then num_positive_labels::float/num_labeled_examples else 0 end as label_base_rate
            from triage_metadata.experiment_matrices join test_results.evaluations using(matrix_uuid)
            where experiment_hash in ('{"','".join(self.experiment_hashes)}') and subset_hash = ''
            order by 1
        """

        matrices = pd.read_sql(q, self.engine)

        print(matrices)
    

    def subset_summary(self, subset_hash):
        q = f"""
            select distinct on(train_end_time)
                -- matrix_uuid,
                evaluation_start_time as train_end_time,
                num_labeled_examples as cohort_size,
                num_positive_labels,
                case when num_labeled_examples > 0 then num_positive_labels::float/num_labeled_examples else 0 end as label_base_rate
            from triage_metadata.experiment_matrices join test_results.evaluations using(matrix_uuid)
            where experiment_hash in ('{"','".join(self.experiment_hashes)}') 
                and subset_hash = '{subset_hash}'
            order by 1
        """

        matrices = pd.read_sql(q, self.engine)

        print(matrices)

    def plot_model_group_performance(self, metric, parameter):
        pass        

        # q = """
        #     select
        
        # """

    def get_model_ids(self):
        """ For the model group ids, fetch the model_ids and initialize the datastructure

            The data structure is a dictionary of dictionaries that maps the individual models of groups to their ModelAnalyzer class
                {model_group_id: {train_end_time: ModelAnalyzer(model_id)}}
                Here, the train_end_time is saved as a string of with format 'YYYY-MM-DD'
        """

        model_groups = "', '".join([str(x) for x in self.model_groups])
        experiment_hashes = "', '".join(self.experiment_hashes)
        q = f"""
            select distinct on (model_group_id, train_end_time)
                model_id, 
                train_end_time::date,
                model_group_id
            from triage_metadata.models 
                join triage_metadata.experiment_models using(model_hash)
            where experiment_hash in ('{experiment_hashes}')
            and model_group_id in ('{model_groups}')        
            """  
        # TODO do we really need experiment_hashes here? can we query with only model_group_ids?
        
        # TODO: modify to remove pandas
        models = pd.read_sql(q, self.engine).to_dict(orient='records')

        d = dict()
        for experiment_hash in self.experiment_hashes:
            q = f"""
                select distinct on (model_group_id, train_end_time)
                    model_id, 
                    to_char(train_end_time::date, 'YYYY-MM-DD') as train_end_time,
                    model_group_id
                from triage_metadata.models 
                    join triage_metadata.experiment_models using(model_hash)
                where experiment_hash='{experiment_hash}'
                and model_group_id in ('{model_groups}')        
                """  

            # TODO: modify to remove pandas
            models = pd.read_sql(q, self.engine).to_dict(orient='records')

            for m in models:
                if m['model_group_id'] in d:
                    d[m['model_group_id']][m['train_end_time']] = SingleModelAnalyzer(m['model_id'], self.engine)
                else:
                    d[m['model_group_id']] = {m['train_end_time']: SingleModelAnalyzer(m['model_id'], self.engine)}

        return d 
    
    def _get_subplots(self, subplot_width=3, subplot_len=None, n_rows=None, n_cols=None, sharey=False, sharex=False):
        """"""

        if subplot_len is None:
            subplot_len = subplot_width

        # num of rows == number of train_end_times
        if n_rows is None: 
            n_rows = len(self.models[self.model_groups[0]])
        
        # num of cols == number of model groups
        if n_cols is None: 
            n_cols = len(self.model_groups)
        
        fig, axes = plt.subplots(
            n_rows,
            n_cols,
            figsize = (subplot_width*n_cols, subplot_len*n_rows),
            dpi=100,
            sharey=sharey, 
            sharex=sharex
        )

        return fig, axes


    def _make_plot_grid(self, plot_type, subplot_width=3, subplot_len=None, sharey=False, sharex=False, **kw):
        """
            Abstracts out generating the plot grid (time x model group) for comparing
            model_scores,  
        """
        fig, axes = self._get_subplots(subplot_width=subplot_width, subplot_len=subplot_len, sharey=sharey, sharex=sharex)
        
        logging.info(f"{len(axes), len(axes[0])}")

        for j, mg in enumerate(self.models):
            for i, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                plot_func = getattr(model_analyzer, plot_type)

                if (len(self.models) == 1) or (len(self.models[mg]) == 1):
                    if (len(self.models) == 1) and (len(self.models[mg]) == 1):
                        ax = axes
                    else:
                        ax = axes[i]
                else:
                    ax = axes[i, j]
                ax = plot_func(ax=ax, **kw)

                if j==0:
                    ax.set_ylabel(f'{train_end_time}')
                else:
                    ax.set_ylabel('')
                
                ax.set_xlabel('')

        fig.tight_layout()

        return fig

    def plot_score_distributions(self, use_labels=False):
        """for the model group ids plot score grid"""

        if use_labels:
            self._make_plot_grid(plot_type='plot_score_label_distribution')
        else:
            self._make_plot_grid(plot_type='plot_score_distribution')
        
    def plot_calibration_curves(self):
        """calibration curves for all models"""
        self._make_plot_grid(plot_type='plot_calibration_curve')

    def plot_prk_curves(self, **kw):
        self._make_plot_grid(plot_type='plot_precision_recall_curve', **kw)
        
    def plot_recall_curves_overlaid(self, n_splits=None, **kw,):
        # Number of columns    
        if n_splits is None:
            n_cols = len(self.models[self.model_groups[0]])
        else:
            n_cols = n_splits
            
        fig, axes = self._get_subplots(
            subplot_width=3, 
            subplot_len=3, 
            sharey=True, 
            sharex=True, 
            n_cols=n_cols, 
            n_rows=1
        )
        
        as_of_dates = list(sorted(self.models[self.model_groups[0]].keys()))[-n_cols:]
        
        for i, aod in enumerate(as_of_dates):
            for model_group in sorted(self.model_groups):
                mod = self.models[model_group].get(aod)
                if mod is None:
                    continue

                mod.plot_precision_recall_curve(only_recall=True,ax = axes[i], title_string=aod)
                axes[i].set(alpha=0.1)
                axes[i].legend().remove()
                axes[i].set_xlabel('Population pct (k%)')
                axes[i].set_ylabel('recall@k')
                sns.despine()
        plt.tight_layout()
        axes[-1].legend(sorted(self.model_groups))
        
        
        

  
    def plot_bias_threshold(self, attribute_name, attribute_values, bias_metric):
        """
            Plot bias_metric for the specified list of attribute_values for a particular attribute_name across different thresholds (list %)
        """
        fig, axes = self._get_subplots(subplot_width=6, n_rows=len(attribute_values), n_cols=len(self.models[self.model_groups[0]]))
        for _, mg in enumerate(self.models):
            for i, attribute_value in enumerate(attribute_values):
                for j, train_end_time in enumerate(self.models[mg]):
                    mode_analyzer = self.models[mg][train_end_time]

                    mode_analyzer.plot_bias_threshold_curve(
                        attribute_name=attribute_name,
                        attribute_value=attribute_value,
                        bias_metric=bias_metric,
                        ax=axes[i, j]
                    )
                    if j==0:
                        axes[i, j].set_ylabel(f'{attribute_name}:{attribute_value}')
                    else:
                        axes[i, j].set_ylabel('')
                    if i == 0:
                        axes[i, j].set_title(f'{train_end_time}')
                    else:
                        axes[i, j].set_title('')
        fig.suptitle(f"{bias_metric} Threshold Curve for {attribute_name}")
        fig.tight_layout()

    def plot_precision_threshold(self):
        """
            Plot precision against threshold (list %)
        """
        if len(self.models) <= 1:
            print("Not available when there is only one model group (look at plot_prk_curves instead)")
            return
        fig, axes = self._get_subplots(subplot_width=6, n_cols=1, sharey=True)
        for _, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                mode_analyzer = self.models[mg][train_end_time]
                mode_analyzer.plot_precision_threshold_curve(
                    ax=axes[j]
                )
                axes[j].set_title(f'{train_end_time}')
        fig.suptitle("Precision Threshold Curve")
        fig.tight_layout()

    def plot_feature_importance(self, n_top_features=20):
        """ plot all feature importance  """
        self._make_plot_grid(plot_type='plot_feature_importance', subplot_width=7, n_top_features=n_top_features)
        
            
    def plot_feature_group_importance(self, n_top_groups=20):
        """ plot all feature group importance  """
        self._make_plot_grid(plot_type='plot_feature_group_importance', subplot_width=7, n_top_groups=n_top_groups)


    def create_scatter_disparity_performance(self, metric, parameter, aeq_parameter, attr_col, attribute_values, 
                                         performance_col='stochastic_value', bias_metric='tpr', tiebreaker='worst', flip_disparity=False, 
                                         mitigated_tags=[], mitigated_bdfs=[], mitigated_performances=[], ylim=None):
        """
            Create scatterplot of one bias metric (e.g. tpr disparity) vs an evaluation metric (e.g. precision at some threshold) for a particular list of attribute value (must be from the same attribute)
            A simplified version of the scatterplot function here: https://github.com/dssg/fairness_tutorial/blob/master/notebooks/bias_reduction_with_outputs.ipynb 
        """
        # TODO add legend for model identification?
        evals_df_list = {}
        aequitas_df_list = {}
        fig, axes = self._get_subplots(subplot_width=6, n_rows=len(attribute_values))
        
        for k, attribute_value in enumerate(attribute_values):
            for i, mg in enumerate(self.models):
                for j, train_end_time in enumerate(self.models[mg]):
                    mode_analyzer = self.models[mg][train_end_time]
                    if i == 0:
                        evals_df_list[train_end_time] = []
                        aequitas_df_list[train_end_time] = []
                    evals_df_list[train_end_time].append(mode_analyzer.get_evaluations(metrics={metric: [parameter]})) # TODO only allow one eval metric here?
                    aequitas_df_list[train_end_time].append(mode_analyzer.get_aequitas())

            for j, train_end_time in enumerate(self.models[self.model_groups[0]]):

                evals_df = pd.concat(evals_df_list[train_end_time])
                aequitas_df = pd.concat(aequitas_df_list[train_end_time])
                # filter aequitas by eval metric
                aequitas_df = aequitas_df.loc[(aequitas_df['parameter']==aeq_parameter) * (aequitas_df['tie_breaker']==tiebreaker)]
                disparity_df = aequitas_df.loc[(aequitas_df['attribute_name']==attr_col) & (aequitas_df['attribute_value']==attribute_value)].copy()
                disparity_metric = bias_metric + '_disparity'
                scatter_schema = ['model_id', performance_col, 'attribute_name', 'attribute_value', bias_metric, disparity_metric, 'model_tag']
                if flip_disparity:
                    disparity_df[disparity_metric]= disparity_df.apply(lambda x: 1/x[disparity_metric] , axis=1)
                scatter = pd.merge(evals_df, disparity_df, how='left', on=['model_id'], sort=True, copy=True)
                scatter = scatter[['model_id', performance_col, 'attribute_name', 'attribute_value', bias_metric, disparity_metric]].copy()
                scatter['model_tag'] = 'Other Models'
                scatter.sort_values('stochastic_value', ascending = False, inplace=True, ignore_index=True)
                scatter_final = pd.DataFrame()

                ax = axes[k,j]
                ax.scatter(
                    x='stochastic_value', y=disparity_metric,
                    data=scatter

                )
                if not scatter_final.empty:
                    ax.scatter(
                        x='stochastic_value', y=disparity_metric,
                        data=scatter_final
                    )
                
                if j==0:
                    axes[k, j].set_ylabel(f'{attr_col}:{attribute_value}')
                else:
                    axes[k, j].set_ylabel('')
                if k == 0:
                    axes[k, j].set_title(f'{train_end_time}')
                else:
                    axes[k, j].set_title('')
                axes[k, j].set_xlabel(f'{metric}{parameter}')

                if ylim:
                    plt.ylim(0, 10)
        flip_placeholder = 'Flipped' if flip_disparity else ''
        fig.suptitle(f'{flip_placeholder} {disparity_metric} vs. {metric}{parameter} for {attr_col}')


    def calculate_crosstabs_pos_vs_neg(self, project_path, thresholds, table_name='crosstabs', **kwargs):
        """ Generate crosstabs for the predicted positives (top-k) vs the rest

        args:
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
            table_name (str, optional): Table name to use in the db's `test_results` schema. Defaults to crosstabs.
                                        If the table exists, results are appended
            **kwargs: This method can take other arguments sent to ModelAnalyzer.crosstabs_pos_vs_ng function
        """

        for i, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                df = model_analyzer.crosstabs_pos_vs_neg(
                    project_path=project_path,
                    thresholds=thresholds,
                    table_name=table_name,
                    **kwargs
                )


    def display_crosstab_pos_vs_neg(
        self, 
        threshold_type,
        threshold, 
        table_name='crosstabs', 
        project_path=None,
        display_n_features=40,
        filter_features=None,
        support_threshold=0.1,
        show_plots=True,
        return_dfs=True):

        """ display crosstabs for one threshold for all the models in the model groups
            
        Args:
            threshold_type (str): Type of rank threshold to use in splitting predicted positives from negatives. 
                                Has to be one of the rank columns in the test_results.predictions_table

            threshold (Union[int, float]): The rank threshold of the specified type. If the threshold type is an absolute, integer. If percentage, should be a float between 0 and 1
                        
            table_name (str, optional): Table name to fetch from/write to in the db's `test_results` schema. Defaults to crosstabs.
            
            project_path (str, optional): Path where the experiment artifacts are stored. Optional if the crosstabs are already calculated,
                required if the crosstab need to be calculated
            
            display_n_features (int, optional): Number of features to return. defaults to 40 (sorted by mean ratio desc). This is ignored if `filter_features` is specified

            filter_features (List[str], optional): The list of features that we are interested in. If not specified, `display_n_features` features are returned

            support_threshold (float, optional): The threshold of pct support for the feature (instances with non-zero values) among predicted positives 
        """

        dfs = dict()

        for i, mg in enumerate(self.models):
            for j, train_end_time in enumerate(self.models[mg]):
                model_analyzer = self.models[mg][train_end_time]

                idx = i * len(self.models[self.model_groups[0]]) + j 


                if len(self.models) == 1 and len(self.models[mg]) == 1:
                    axes = [axes] 
                try:
                    df = model_analyzer.display_crosstabs_pos_vs_neg(
                        threshold_type=threshold_type,
                        threshold=threshold,
                        display_n_features=display_n_features,
                        filter_features=filter_features,
                        support_threshold=support_threshold,
                        table_name=table_name,
                        # ax=(axes if show_plots else None),
                        show_plot=False, #TODO -- remove this parameter from 
                        return_df=True
                    )

                    dfs[model_analyzer.model_id] = df
                except ValueError as e:
                    logging.error('Please run calculate_crosstabs_pos_vs_neg function to calculate crosstabs first for all models!')
                    raise e
                
                print(f'\nModel Group: {mg}, Validation date: {train_end_time}'.center(30, ' '))
                display(df)

        if return_dfs:
            return dfs
        
    def _pairwise_feature_importance_comparison_single_split(self, train_end_time, n_top_features, model_group_ids=None,plot=True):
        """ For a given train_end_time, compares the top n features (highest absolute importance) of two models 
            
            Args:
                train_end_time (str): The prediction date we care about in YYYY-MM-DD format
                n_top_features (int): Number of features to consider for the comparispn
                model_group_ids (int, optional): Model group ids to consider, if not provided, all model groups included in the report are used
                plot (bool, optional): Whether to plot the results. Defaults to True.
        """
        
        feature_lists = dict()
        
        if model_group_ids is not None:
            
            for mg in model_group_ids:
                feature_lists[mg] = self.models[mg][train_end_time].get_feature_importances(n_top_features=n_top_features) 
                if feature_lists[mg].empty:
                    logging.warning('No feature importance values were found for model group {mg}. Excluding from comparison')
                    feature_lists.pop(mg)
        # By default all feature importance values are considered
        else:
            model_group_ids = self.model_groups
            for mg, m in self.models.items():
                feature_lists[mg] = m[train_end_time].get_feature_importances(n_top_features=n_top_features) 
                if feature_lists[mg].empty:
                    logging.warning('No feature importance values were found for model group {mg}. Excluding from comparison')
                    feature_lists.pop(mg)
            
        pairs = list(itertools.combinations(feature_lists.keys(), 2))
        
        logging.info(f'Performing {len(pairs)} comparisons')
        
        metrics = ['jaccard', 'overlap', 'rank_corr']
        results = dict()
        
        for m in metrics:
            results[m] = pd.DataFrame(index=sorted(model_group_ids), columns=sorted(model_group_ids))
            # filling the diagonal with 1
            results[m].values[[np.arange(results[m].shape[0])]*2] = 1
            
        for model_group_pair in pairs:
            model_group_pair = sorted(model_group_pair)
            logging.info(f'Comparing {model_group_pair[0]} and {model_group_pair[1]}')

            df1 = feature_lists[model_group_pair[0]]
            df2 = feature_lists[model_group_pair[1]]
            
            f1 = set(df1.feature)
            f2 = set(df2.feature)
            
            if (len(f1) == 0 or len(f2)) == 0:
                logging.error('No feature importance available for the models!') 
                continue

            inter = f1.intersection(f2)
            un = f1.union(f2)    
            results['jaccard'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/len(un)

            # If the list sizes are not equal, using the smallest list size to calculate simple overlap
            results['overlap'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/ min(len(f1), len(f2))

            # calculating rank correlation
            df1.sort_values('feature_importance', ascending=False, inplace=True)
            df2.sort_values('feature_importance', ascending=False, inplace=True)

            # only returning the corr coefficient, not the p-value
            results['rank_corr'].loc[model_group_pair[0], model_group_pair[1]] = spearmanr(df1.feature.iloc[:], df2.feature.iloc[:])[0]

    
        if plot:
            fig, axes = plt.subplots(1, len(metrics), figsize=(10, 3))            
            
            for i, m in enumerate(metrics):
                sns.heatmap(
                    data=results[m].fillna(0),
                    cmap='Greens',
                    vmin=0,
                    vmax=1,
                    annot=True,
                    linewidth=0.1,
                    ax=axes[i]
                )

                axes[i].set_title(m)

            fig.suptitle(train_end_time)
            fig.tight_layout()
        
        return results
    

    def _pairwise_list_comparison_single_fold(self, threshold_type, threshold, train_end_time, matrix_uuid=None, plot=True):
        """For a given train_end_time, compares the lists generated by the analyzed model groups

        Args:
            threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                    - rank_pct_no_ties 
                    - rank_pct_with_ties
                    - rank_abs_no_ties
                    - rank_abs_with_ties
            threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
            train_end_time (str): The prediction date we care about in YYYY-MM-DD format
        """

        lists = dict()

        for mg, m in self.models.items():
            lists[mg] = m[train_end_time].get_top_k(threshold_type, threshold, matrix_uuid)

        
        pairs = list(itertools.combinations(lists.keys(), 2))

        logging.info(f'Performing {len(pairs)} comparisons')

        metrics = ['jaccard', 'overlap', 'rank_corr']
        results = dict()

        # Initializing three data frames to hold pairwise metrics
        for m in metrics:
            results[m] = pd.DataFrame(index=sorted(self.model_groups), columns=sorted(self.model_groups))
            results[m].values[[np.arange(results[m].shape[0])]*2] = 1

        for model_group_pair in pairs:
            logging.info(f'Comparing {model_group_pair[0]} and {model_group_pair[1]}')

            model_group_pair = sorted(model_group_pair)
            
            df1 = lists[model_group_pair[0]]
            df2 = lists[model_group_pair[1]]

            # calculating jaccard similarity and overlap
            entities_1 = set(df1.entity_id)
            entities_2 = set(df2.entity_id)

            if (len(entities_1) == 0 or len(entities_2)) == 0:
                logging.error('No prediction saved for the models!') 

            inter = entities_1.intersection(entities_2)
            un = entities_1.union(entities_2)    
            results['jaccard'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/len(un)

            # If the list sizes are not equal, using the smallest list size to calculate simple overlap
            results['overlap'].loc[model_group_pair[1], model_group_pair[0]] = len(inter)/ min(len(entities_1), len(entities_2))

            # calculating rank correlation
            df1.sort_values('score', ascending=False, inplace=True)
            df2.sort_values('score', ascending=False, inplace=True)

            # only returning the corr coefficient, not the p-value
            if len(df1) == len(df2):
                results['rank_corr'].loc[model_group_pair[0], model_group_pair[1]] = spearmanr(df1.entity_id.iloc[:], df2.entity_id.iloc[:])[0]
            else:
                logging.warning(f'Not calculating rank correlation. List sizes are not equal ({len(df1)}, {len(df2)})')

        if plot:
            fig, axes = plt.subplots(1, len(metrics), figsize=(10, 3))            
            
            for i, m in enumerate(metrics):
                sns.heatmap(
                    data=results[m].fillna(0),
                    cmap='Greens',
                    vmin=0,
                    vmax=1,
                    annot=True,
                    linewidth=0.1,
                    ax=axes[i]
                )

                axes[i].set_title(m)

            fig.suptitle(train_end_time)
            fig.tight_layout()
        return results

    def _get_individual_model_ids(self):
        """Get individual model ids associated with the given model groups
        """
        model_groups = ", ".join([str(x) for x in self.model_groups])

        q = f"""
            select
                model_id 
            from triage_metadata.models
            where model_group_id in({model_groups})
        """
        df_model_ids = pd.read_sql(q, self.engine)

        return df_model_ids.model_id.to_list()


    def execute_error_analysis(self, model_ids=None):
        """Generates the error analysis of a model
            args:
                model_ids (int, optional): A list of model_ids we are interested in analyzing. If not provided, all models are considered.
        """
        if model_ids is None:
            model_ids = self.model_ids

        for model_id in model_ids:
            logging.info(f"generating error analysis for model id {model_id}")
            results = generate_error_analysis(model_id, self.engine, self.project_path)
            output_all_analysis(results)
    
    
    def pairwise_top_k_list_comparison(self, threshold_type, threshold, train_end_times=None, matrix_uuid=None, plot=True):
        """
            Compare the top-k lists for the given train_end_times for all model groups considered (pairwise)
            
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                        - rank_pct_no_ties 
                        - rank_pct_with_ties
                        - rank_abs_no_ties
                        - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                train_end_times (Optional, List[str]): The prediction date we care about in YYYY-MM-DD format
        """

        # If no train_end_times are provided, we consider all the train_end_times
        # NOTE -- Assuming that the all model groups have the same train_end_times
        if train_end_times is None:
            train_end_times = self.models[self.model_groups[0]].keys()

        for train_end_time in train_end_times:
            self._pairwise_list_comparison_single_fold(
                threshold=threshold,
                threshold_type=threshold_type,
                train_end_time=train_end_time,
                matrix_uuid=matrix_uuid,
                plot=plot
            )
            
    def pairwise_feature_importance_comparison(self, n_top_features, model_groups=None, train_end_times=None, plot=True):
        """
            Compare the top-k lists for the given train_end_times for all model groups considered (pairwise)
            
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                        - rank_pct_no_ties 
                        - rank_pct_with_ties
                        - rank_abs_no_ties
                        - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                train_end_times (Optional, List[str]): The prediction date we care about in YYYY-MM-DD format
        """

        # If no train_end_times are provided, we consider all the train_end_times
        # NOTE -- Assuming that the all model groups have the same train_end_times
        if train_end_times is None:
            train_end_times = self.models[self.model_groups[0]].keys()
            

        for train_end_time in train_end_times:
            self._pairwise_feature_importance_comparison_single_split(
                train_end_time=train_end_time,
                n_top_features=n_top_features,
                model_group_ids=model_groups,
                plot=plot
            )



import ohio.ext.pandas
import pandas as pd
import numpy as np
import logging
import seaborn as sns
import matplotlib.table as tab
import matplotlib.pyplot as plt
from tabulate import tabulate

from descriptors import cachedproperty
from sqlalchemy import create_engine
from sklearn.calibration import calibration_curve
from sklearn import metrics

from triage.component.catwalk.storage import ProjectStorage
from triage.component.postmodeling.error_analysis import generate_error_analysis
from triage.database_reflection import table_exists
from triage.component.catwalk.utils import sort_predictions_and_labels

class ModelAnalyzer:

    id_columns = ['entity_id', 'as_of_date']

    def __init__(self, model_id, engine):
        self.model_id=model_id
        self.engine=engine

    @cachedproperty
    def metadata(self):
        return next(self.engine.execute(
                    f'''
                    WITH individual_model_ids_metadata AS(
                        SELECT m.model_id,
                           m.model_group_id,
                           m.hyperparameters,
                           m.model_hash,
                           m.train_end_time::date,
                           m.train_matrix_uuid,
                           m.training_label_timespan,
                           m.model_type,
                           mg.model_config
                        FROM triage_metadata.models m
                        JOIN triage_metadata.model_groups mg
                        USING (model_group_id)
                        WHERE model_id = {self.model_id}
                    ),
                    individual_model_id_matrices AS(
                        SELECT DISTINCT ON (matrix_uuid)
                           model_id,
                           matrix_uuid,
                           evaluation_start_time as as_of_date
                        FROM test_results.evaluations
                        WHERE model_id = ANY(
                            SELECT model_id
                            FROM individual_model_ids_metadata
                        )
                    )
                    SELECT metadata.*, test.*
                    FROM individual_model_ids_metadata AS metadata
                    LEFT JOIN individual_model_id_matrices AS test
                    USING(model_id);'''
            )
        )
    
    @property
    def model_group_id(self):
        return self.metadata['model_group_id']

    @property
    def model_type(self):
        return self.metadata['model_type']

    @property
    def hyperparameters(self):
        return self.metadata['hyperparameters']

    @property
    def model_hash(self):
        return self.metadata['model_hash']

    @property
    def train_matrix_uuid(self):
        return self.metadata['train_matrix_uuid']

    # TODO: Need to figure out how this would work when there are multiple matrices in the evaluations table
    @property
    def pred_matrix_uuid(self):
        return self.metadata['matrix_uuid']

    @property
    def as_of_date(self):
        return self.metadata['as_of_date']

    @property
    def train_end_time(self):
        return self.metadata['train_end_time']

    @property
    def train_label_timespan(self):
        return self.metadata['training_label_timespan']

    def get_predictions(self, matrix_uuid=None, fetch_null_labels=True, predictions_table='test_results.predictions', subset_hash=None):
        """Fetch the predictions from the DB for a given matrix
        
            args:
                matrix_uuid (optional):  
        
        """
        where_clause = f"WHERE model_id = {self.model_id}"

        if matrix_uuid is not None:
            where_clause += f" AND matrix_uuid='{matrix_uuid}'"

        if not fetch_null_labels:
            where_clause += f" AND label_value IS NOT NULL"

        if subset_hash:
            # get subset table name
            q = f"select config from triage_metadata.subsets where subset_hash='{subset_hash}'"
            config_df = pd.read_sql(q, self.engine)
            table_name = f"subset_{config_df.iloc[0]['config']['name']}_{subset_hash}"
            predictions_table += f" preds join {table_name} subset on preds.entity_id = subset.entity_id and preds.as_of_date = subset.as_of_date"

        query = f"""
            SELECT model_id,
                   entity_id,
                   as_of_date,
                   score,
                   label_value,
                   rank_abs_with_ties,
                   rank_pct_with_ties,
                   rank_abs_no_ties,
                   rank_pct_no_ties, 
                   test_label_timespan
            FROM {predictions_table}
            {where_clause}        
        """

        preds = pd.read_sql(query, self.engine)

        #TODO: Maybe we should call the script to save predictions here?
        if preds.empty:
            logging.warning(f'No predictions were found in {predictions_table} for model_id {self.model_id}. Returning empty dataframe!')
            # raise RuntimeError(
                # "No predictions were found in the database. Please run the add_predictions module to add predictions for the model"
            # )
            return preds 
        

        return preds.set_index(self.id_columns)

    
    def get_top_k(self, threshold_type, threshold, matrix_uuid=None):
        """ Fetch the k intities with highest model score for a given model
        
            Args:
                threshold_type (str): Type of the ranking to use. Has to be one of the four ranking types used in triage
                    - rank_pct_no_ties 
                    - rank_pct_with_ties
                    - rank_abs_no_ties
                    - rank_abs_with_ties
                threshold (Union[float, int]): The threshold rank for creating the list. Int for 'rank_abs_*' and Float for 'rank_pct_*'
                matrix_uuid (optional, str): If a list should be generated out of a matrix that were not used to validate the model during the experiment
        """

        q = f"""
            select 
                entity_id, 
                as_of_date,
                score,
                label_value,
                rank_abs_no_ties,
                rank_abs_with_ties,
                rank_pct_no_ties,
                rank_pct_with_ties,
                matrix_uuid
            from test_results.predictions
            where model_id={self.model_id}
            and {threshold_type} <= {threshold}
        """

        top_k = pd.read_sql(q, self.engine)

        return top_k


    # TODO generalize this function
    # TODO do we want to allow specifying a dictionary of attributes + values? 
    def get_aequitas(self, parameter=None, attribute_name=None, subset_hash=None):
        '''
        Get aequitas evaluations from the DB

        Args:
            parameter (str): Optional. The threshold to apply when returning the aequitas evaluations.
                                  If not specified, all aequitas evaluations will be returned.

            attribute_name (str): Optional. Fetch aequitas evaluations related to a particular attribute.

            subset_hash (str): Optional. For fetching evaluations of a specific subset.    
        '''

        where_clause = f'WHERE model_id={self.model_id}'

        if subset_hash is not None:
            where_clause += f" AND 'subset_hash='{subset_hash}'"
        else:
            where_clause += f" AND subset_hash=''"

        if parameter is not None:
            where_clause += f" AND parameter='{parameter}'"

        if attribute_name:
            where_clause += f" AND attribute_name='{attribute_name}'"
        
        # TODO don't return all columns ?
        q = f"""
            select
                * 
            from test_results.aequitas
            {where_clause}
        """

        evaluations = pd.read_sql(q, self.engine)

        return evaluations
    

    def get_evaluations(self, metrics=None, matrix_uuid=None, subset_hash=None):
        ''' 
        Get evaluations for the model from the DB

        Args:
            metrics Dict[str:List]): Optional. The metrics and parameters for evaluations. 
                                    A dictionary of type {metric:[thresholds]}
                                    If not specified, all the evaluations will be returned

            matrix_uuid (str): Optional. If model was evaluated using multiple matrices
                            one could get evaluations of a specific matrix. Defaults to fetching everything

            subset_hash (str): Optional. For fetching evaluations of a specific subset.    
        '''

        where_clause = f'WHERE model_id={self.model_id}'

        if matrix_uuid is not None:
            where_clause += f" AND matrix_uuid='{matrix_uuid}'"

        if subset_hash is not None:
            where_clause += f" AND subset_hash='{subset_hash}'"
        else:
            where_clause += f" AND subset_hash=''"

        if metrics is not None:
            where_clause += " AND ("
            for i, metric in enumerate(metrics):
                parameters = metrics[metric]
                where_clause += f""" metric='{metric}' AND parameter in ('{"','".join(parameters)}')"""

                if i < len(metrics) - 1:
                    where_clause += "OR"

            where_clause += ") "

        q = f"""
            select
                model_id,
                matrix_uuid,
                subset_hash,
                metric, 
                parameter,
                stochastic_value,
                num_labeled_above_threshold,               
                num_positive_labels
            from test_results.evaluations
            {where_clause}
        """

        evaluations = pd.read_sql(q, self.engine)

        return evaluations

    def get_feature_importances(self, n_top_features=20):

        logging.debug(f'Fetching feature importance from db for model id: {self.model_id}')
        features = pd.read_sql(
           f'''
           select
                feature,
                feature_importance,
                rank_abs
           FROM train_results.feature_importances
           WHERE model_id = {self.model_id}
           and rank_abs <= {n_top_features}
           and abs(feature_importance) > 0 
           order by rank_abs
           ''', con=self.engine)
        return features
    
    def get_feature_group_importances(self):
        """
        Get the top most important feature groups as identified by the maximum importance of any feature in the group

        """
        # TODO this assumes any experiment linked to this model has the same feature aggregations, is this valid?
        q = f"""
            select distinct experiment_hash from triage_metadata.models m 
            left join triage_metadata.experiment_models em on m.model_hash=em.model_hash 
            where model_group_id={self.model_group_id}
            """
        experiment_hashes = pd.read_sql(q, self.engine)
        experiment_hash = experiment_hashes['experiment_hash'].iloc[0]

        # get feature group names
        q = f"""
            select 
                config->'feature_aggregations' as feature_groups
            from triage_metadata.experiments where experiment_hash = '{experiment_hash}'
        """

        feature_groups = [i['prefix'] for i in pd.read_sql(q, self.engine)['feature_groups'].iloc[0]]
        feature_groups
        case_part = ''
        for fg in feature_groups:
            case_part = case_part + "\nWHEN feature like '{fg}%%' THEN '{fg}'".format(fg=fg)

        # get feature group importances
        feature_group_importance = pd.read_sql(f"""
            with raw_importances as (
                select 
                    model_id,
                    feature,
                    feature_importance,
                    CASE {case_part}
                    ELSE 'No feature group'
                    END as feature_group
                FROM train_results.feature_importances
                WHERE model_id = {self.model_id}
            )
            SELECT
            model_id,
            feature_group,
                max(abs(feature_importance)) as importance_aggregate
            FROM raw_importances
            GROUP BY feature_group, model_id
        """, con=self.engine)
        return feature_group_importance

    def crosstabs_pos_vs_neg(self, project_path, thresholds, matrix_uuid=None, push_to_db=True, table_name='crosstabs', return_df=True, replace=True, predictions_table='test_results.predictions'):
        """ Generate crosstabs for the predicted positives (top-k) vs the rest
    
        args:
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
            return_df (bool, optional): Whether to return the constructed df or just to store in the database
                                        Defaults to False (only writing to the db)
            
            matrix_uuid (str, optional): To run crosstabs for a different matrix than the validation matrix from the experiment

            push_to_db (bool, optional): Whether to write the results to the database. Defaults to True
            table_name (str, optional): Table name to use in the db's `test_results` schema. Defaults to crosstabs.
                                        If the table exists, results are appended

            return_df (bool, optional): Whether to return the crosstabs as a dataframe. Defaults to True
        """

        # metrics we calculate for positive and negative predictions
        positive_mean = lambda pos, neg: pos.mean(axis=0)
        negative_mean = lambda pos, neg: neg.mean(axis=0)
        positive_std = lambda pos, neg: pos.std(axis=0)
        negative_std = lambda pos, neg: neg.std(axis=0)
        ratio_positive_negative = lambda pos, neg: pos.mean(axis=0) / neg.mean(axis=0)
        # mean_ratio = lambda pos, neg: max(ratio_positive_negative(pos, neg).at[0], (float(1) / ratio_positive_negative(pos, neg)).at[0])
        positive_support = lambda pos, neg: (pos > 0).sum(axis=0)
        negative_support = lambda pos, neg: (neg > 0).sum(axis=0)
        positive_support_pct = lambda pos, neg: round((pos > 0).sum(axis=0).astype(float) / len(pos), 3)
        negative_support_pct = lambda pos, neg: round((neg > 0).sum(axis=0).astype(float) / len(neg), 3)

        def mean_ratio(pos, neg):
            pos_over_neg = ratio_positive_negative(pos, neg)
            neg_over_pos = float(1) / pos_over_neg

            df = pd.DataFrame([pos_over_neg, neg_over_pos])

            return df.max()

        crosstab_functions = [
            ("mean_predicted_positive", positive_mean),
            ("mean_predicted_negative", negative_mean),
            ("std_predicted_positive", positive_std),
            ("std_predicted_negative", negative_std),
            ("mean_ratio", mean_ratio),
            ("support_predicted_positive", positive_support),
            ("support_predicted_negative", negative_support),
            ("support_pct_predicted_positive", positive_support_pct),
            ("support_pct_predicted_negative", negative_support_pct)
        ]


        if matrix_uuid is None:
            matrix_uuid = self.pred_matrix_uuid
            logging.debug(f'Matrix uuid set to: {matrix_uuid}')

        predictions = self.get_predictions(matrix_uuid=matrix_uuid, predictions_table=predictions_table)

    
        if predictions.empty:
            logging.error(f'No predictions found for {self.model_id} and {matrix_uuid}. Exiting!')
            raise ValueError(f'No predictions found {self.model_id} and {matrix_uuid}')

        # Check whether the table exists
        if table_exists(f'test_results.{table_name}', self.engine):
            # checking whether the crosstabs already exist for the model
            logging.debug(f'Checking whether crosstabs already exist for the model {self.model_id} and {matrix_uuid}')
            q = f"select * from test_results.{table_name} where model_id={self.model_id} and matrix_uuid='{matrix_uuid}';"
            df = pd.read_sql(q, self.engine)
            
            if not df.empty:
                logging.warning(f'Crosstabs aleady exist for model {self.model_id} and matrix_uuid={matrix_uuid}')

                if replace:
                    logging.warning('Deleting the existing crosstabs!')
                    with self.engine.connect() as conn:
                        conn.execute(f"delete from test_results.{table_name} where model_id={self.model_id} and matrix_uuid='{matrix_uuid}';")
                else:
                    logging.info(f"Replace set to False. Not calculating crosstabs for model {self.model_id} and matrix_uuid='{matrix_uuid}';")
                    if return_df: return df 
                    else: return 
            
        # initializing the storage engines
        project_storage = ProjectStorage(project_path)
        matrix_storage_engine = project_storage.matrix_storage_engine()

        matrix_store = matrix_storage_engine.get_store(matrix_uuid=matrix_uuid)

        matrix = matrix_store.design_matrix
        
        labels = matrix_store.labels
        features = matrix.columns

        # joining the predictions to the model
        matrix = predictions.join(matrix, how='left')

        all_results = list() 
        
        for threshold_name, threshold in thresholds.items():
            logging.info(f'Crosstabs using threshold: {threshold_name} <= {threshold}')

            msk = matrix[threshold_name] <= threshold
            postive_preds = matrix[msk][features]
            negative_preds = matrix[~msk][features]

            temp_results = list()
            for name, func in crosstab_functions:
                logging.info(name)

                this_result = pd.DataFrame(func(postive_preds, negative_preds))
                this_result['metric'] = name
                temp_results.append(this_result)
            
            temp_results = pd.concat(temp_results)
            temp_results['threshold_type'] = threshold_name
            temp_results['threshold'] = threshold

            all_results.append(temp_results)

        crosstabs_df = pd.concat(all_results).reset_index()

        crosstabs_df.rename(columns={'index': 'feature', 0: 'value'}, inplace=True)
        crosstabs_df['model_id'] = self.model_id
        crosstabs_df['matrix_uuid'] = matrix_uuid

    
        if push_to_db:
            logging.info('Pushing the results to the DB')
            crosstabs_df.set_index(
                ['model_id', 'matrix_uuid', 'feature', 'metric', 'threshold_type', 'threshold'], inplace=True
            )

            # TODO: Figure out to change the owner of the table
            crosstabs_df.pg_copy_to(schema='test_results', name=table_name, con=self.engine, if_exists='append')

        if return_df:
            return crosstabs_df
        

    def display_crosstabs_pos_vs_neg(
            self, 
            threshold_type, threshold, table_name='crosstabs',
            display_n_features=40,
            filter_features=None,
            support_threshold=0.1,
            return_df=True,
            show_plot=True,
            matrix_uuid=None,
            ax=None):
        """ Display the crosstabs for the model

        Args:
            threshold_type (str): Type of rank threshold to use in splitting predicted positives from negatives. 
                                Has to be one of the rank columns in the test_results.predictions_table

            threshold (Union[int, float]): The rank threshold of the specified type. If the threshold type is an absolute, integer. 
                                        If percentage, should be a float between 0 and 1
        
            table_name (str, optional): Table name to fetch crosstabs from (`test_results` schema). Defaults to 'crosstabs'.
        """

        q = f"""
            select 
                model_id, 
                feature, 
                metric, 
                abs(value) as value 
            from test_results.{table_name} 
            where threshold_type = '{threshold_type}'
            and threshold = {threshold}
            and metric in (
                'mean_ratio',
                'mean_predicted_positive',
                'mean_predicted_negative',
                'support_pct_predicted_positive',
                'support_pct_predicted_negative'
            )
            and model_id = {self.model_id}            
        """

        if matrix_uuid is not None:
            q = q + f"\n and matrix_uuid='{matrix_uuid}'"
        else:
            matrix_uuid=self.pred_matrix_uuid

        ct = pd.read_sql(q, self.engine)

        if ct.empty:
            logging.error(
                f'''Crosstabs not found for model {self.model_id} and matrix={matrix_uuid} in table test_results.{table_name}. 
                Please use crosstabs_pos_vs_neg function to generate crosstabs'''
            )
            raise ValueError('Crosstabs not found!')

        # Creating the pivot table to convert to wide format indexed by the column
        pivot_table = pd.pivot(
            ct, 
            index='feature', 
            columns='metric', 
            values='value'
        )

        # Shortening the names, and removing the index names to plot more cleanly
        pivot_table.rename(
            columns={
                'mean_ratio': 'ratio',
                'mean_predicted_positive': '(+)mean',
                'mean_predicted_negative': '(-)mean',
                'support_pct_predicted_negative': '(-)supp',
                'support_pct_predicted_positive': '(+)supp'
            }, 
            inplace=True
        )
        pivot_table.index.name='feature_name'
        pivot_table.columns.name=''

        if filter_features is not None:
            return pivot_table.loc[filter_features]

        
        # Filtering by the support threshold
        msk1 = pivot_table['(+)supp'] > support_threshold
        # Features with highest postive : negative ratio
        # df1 = pivot_table[msk].sort_values(
        #     'ratio', 
        #      ascending = False
        # ).head(display_n_features)


        msk2 = pivot_table['(-)supp'] > support_threshold
        # Features with the highest negative : positive ratio
        # df2 = pivot_table[msk].sort_values(
        #     ['ratio', '(-)supp'], 
        #      ascending = [True, False]
        # ).head(display_n_features)

        df = pivot_table[msk1 | msk2].sort_values(
            'ratio', 
             ascending = False
        ).head(display_n_features)

        if show_plot:
            if ax is None: 
                fig, ax = plt.subplots(figsize=(4, 1 + (display_n_features *2) / 5), dpi=100)

            t = tab.table(
                ax, 
                cellText=df.values.round(1), 
                colLabels=df.columns, 
                rowLoc='right',
                rowLabels=df.index, 
                bbox=[0, 0, 1, 1]
            )
        
            ax.set_title(
                f'Model Group: {self.model_group_id} | Train end: {self.train_end_time} | Model: {self.model_id}\n',
                fontdict={
                    'fontsize': 10,
                    'verticalalignment': 'center',
                    'horizontalalignment': 'center'
                },
                x=-0.1
            )
            ax.axis('off')

            # Table formatting
            t.auto_set_column_width([0, 1, 2, 3, 4])
            for key, cell in t.get_celld().items():
                cell.set_fontsize(9)
                cell.set_linewidth(0.1)
                # cell.PAD = 0.01
                
        
        title_str = f'model_group: {self.model_group_id} | train_end_time: {self.train_end_time} | model_id: {self.model_id}'
        # print(title_str)
        # print(tabulate(df.round(1), headers='keys', tablefmt='RST'))


        if return_df:
            df.style.set_caption(title_str)
            return df    
        

    def error_analysis(self, project_path):
        """
        Generates three main error anlaysis with Decision Trees in order to identity 
        what are the features with most importance when the model make mistakes 
        in negative labels (FN), positive labels (FP) and both, negative and 
        positive label errors (FN & FP).

        Args:
            project_path (string): Path for the output of the project 

        Returns: 
            error_analysis_results (list): List of dictionaries with all the error
            analysis made.
        """
        model_id = self.model_id
        connection = self.engine
        error_analysis_results = generate_error_analysis(model_id, connection, project_path=project_path)
        #TODO do we want to 
        return error_analysis_results


    def plot_score_distribution(self, ax, nbins=10, top_k=None, matrix_uuid=None):
        """
        Plot the distribution of predicted scores for all entities across model groups and train_end_time
        Optionally, show only the top k predicted scores.

        Args:
            ax: matplotlib Axes object to plot on
            nbins (optional, int): the number of bins to apply to the score histogram
            top_k (optional, int): if not None, displays only the top_k scores.
            matrix_uuid (optional): get model scores for a particular matrix
        """

        predictions = self.get_predictions(matrix_uuid)
        # keep only top_k if specified
        if top_k:
            predictions = predictions.loc[predictions['rank_abs_no_ties'] <= top_k]
        ax.hist(predictions.score,
                 bins=nbins,
                 alpha=0.5,
                 color='blue')
        ax.axvline(predictions.score.mean(),
                    color='black',
                    linestyle='dashed')
        ax.set_ylabel('Frequency')
        ax.set_xlabel('Score')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')

        return ax

    def plot_score_label_distribution(self, ax,
                                      nbins=10,
                                      top_k=None,
                                      matrix_uuid=None, 
                                      label_names = ('Label = 0', 'Label = 1')):
        """
        Plot the distribution of predicted scores for all entities across model groups and train_end_time
        Optionally, show only the top k predicted scores.
        NOTE: this function only handles 0/1 labels, ignores null labels

        Args:
            ax: matplotlib Axes object to plot on
            nbins (int): the number of bins to define the score distribution
            top_k (optional, int): if not None, displays only the top_k scores.
            matrix_uuid (optional): specify a matrix to get predictions for
            label_names (tuple[String]): specify the two label names to display on the plot

        Return:
            ax: the modified Axes object
        """

        df_predictions = self.get_predictions(matrix_uuid)
        # keep only top_k if specified
        if top_k:
            df_predictions = df_predictions.loc[df_predictions['rank_abs_no_ties'] <= top_k]
        df__0 = df_predictions[df_predictions.label_value == 0]
        df__1 = df_predictions[df_predictions.label_value == 1]

        ax.hist(df__0.score,
                 bins=nbins,
                 alpha=0.5,
                 color='skyblue',
                 label=label_names[0])
        ax.hist(list(df__1.score),
                 bins=nbins,
                 alpha=0.5,
                 color='orange',
                 label=label_names[1])
        ax.axvline(df__0.score.mean(),
                    color='skyblue',
                    linestyle='dashed')
        ax.axvline(df__1.score.mean(),
                    color='orange',
                    linestyle='dashed')
        ax.legend(bbox_to_anchor=(0., 1.005, 1., .102),
                   loc=8,
                   ncol=2,
                   borderaxespad=0.)
        ax.set_ylabel('Frequency')
        ax.set_xlabel('Score')
        ax.set_title('Score Distribution across Labels')
        return ax
    
    # TODO make this more general purpose
    def plot_bias_threshold_curve(self, attribute_name, attribute_value, bias_metric, ax):
        bias_df = self.get_aequitas()
        # limit to specific attribute+value
        bias_df = bias_df.loc[(bias_df['attribute_name']==attribute_name) & (bias_df['attribute_value']==attribute_value)]

        bias_df['perc_points'] = [x.split('_')[0] for x in bias_df['parameter'].tolist()]
        bias_df['perc_points'] = pd.to_numeric(bias_df['perc_points'])

        msk_pct = bias_df['parameter'].str.contains('pct')

        # plot precision
        sns.lineplot(
            x='perc_points',
            y=bias_metric, 
            data=bias_df[msk_pct], 
            label=self.model_group_id,
            ax=ax, 
            estimator='mean', ci='sd'
        )
        ax.set_xlabel('List size percentage (k%)')
        #ax.set_ylabel(f'{bias_metric} for {attribute_name}:{attribute_value}')
        return ax

    def plot_precision_threshold_curve(self, ax, matrix_uuid=None):
        eval_df = self.get_evaluations(matrix_uuid=matrix_uuid)

        eval_df['perc_points'] = [x.split('_')[0] for x in eval_df['parameter'].tolist()]
        eval_df['perc_points'] = pd.to_numeric(eval_df['perc_points'])

        msk_prec = eval_df['metric']=='precision@'
        msk_pct = eval_df['parameter'].str.contains('pct')

        # plot precision
        sns.lineplot(
            x='perc_points',
            y='stochastic_value', 
            data=eval_df[msk_pct & msk_prec], 
            label=self.model_group_id,
            ax=ax, 
            estimator='mean', ci='sd'
        )
        ax.set_xlabel('List size percentage (k%)')
        ax.set_ylabel('Precision')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        return ax

    # def plot_recall_curve(self, ax=None, matrix_uuid=None, list_size_upper_bound_pct=1, pct_step_size=0.01, subset_hash=None, title_string=None):
    #     pred_df = self.get_predictions(matrix_uuid=matrix_uuid)
        
        
    # TODO: Facilitate plotting pr-k with absolute thresholds
    def plot_precision_recall_curve(self, ax=None, matrix_uuid=None, list_size_upper_bound_pct=1, pct_step_size=0.01, subset_hash=None, title_string=None, only_recall=False):
        """
        Plots precision-recall curves at each train_end_time for all model groups
        
        Args:
            ax: matplotlib Axes object to plot on
            matrix_uuid (optional): specify a matrix to get predictions for
            
        Return:
            ax: the modified Axes object
        """

        pred_df = self.get_predictions(matrix_uuid=matrix_uuid)
        # remove null labels
        pred_df.dropna(axis=0, subset=['label_value'], inplace=True) 
        # handle ties
        pred_df_score, pred_df_label, pred_df_index = sort_predictions_and_labels(pred_df['score'], pred_df['label_value'], pred_df.index, tiebreaker="random", sort_seed=15321)
        pred_df = pd.DataFrame(list(zip(pred_df_score, pred_df_label)), columns=['score', 'label_value'])
        
        if ax is None:
            fig, ax = plt.subplots()


        if pred_df.empty or subset_hash is not None:
            logging.warning(f'''
                            No predictions were found for model id {self.model_id} (group: {self.model_group_id}) 
                            or a subset_hash was provided. Using the evaluations to generate the plot. Zoomed PR-K not supported!'''
                            )
            eval_df = self.get_evaluations(matrix_uuid=matrix_uuid, subset_hash=subset_hash)
            
            if eval_df.empty: 
                logging.error('No evaluations were found! Returning empty axes!')
                return ax
            
            eval_df['perc_points'] = [x.split('_')[0] for x in eval_df['parameter'].tolist()]
            eval_df['perc_points'] = pd.to_numeric(eval_df['perc_points'])
            
            msk_prec = eval_df['metric']=='precision@'
            msk_recall = eval_df['metric']=='recall@'
            msk_pct = eval_df['parameter'].str.contains('pct')

            # plot precision
            if not only_recall:
                sns.lineplot(
                    x='perc_points',
                    y='stochastic_value', 
                    data=eval_df[msk_pct & msk_prec], 
                    label='precision@k',
                    ax=ax, 
                    estimator='mean', ci='sd'
                )
            # plot recall
            sns.lineplot(
                x='perc_points', 
                y='stochastic_value', 
                data=eval_df[msk_pct & msk_recall], 
                label='recall@k', 
                ax=ax, 
                estimator='mean', 
                ci='sd'
            )

        else:
            logging.debug(f'Found saved predictions for model id {self.model_id}.(group: {self.model_group_id})')
            k_values = np.arange(0, list_size_upper_bound_pct + pct_step_size, pct_step_size)

            precisions = list()
            recalls = list()
            num_scored = len(pred_df)
            num_positives = pred_df.label_value.sum()

            for k in k_values:
                num_above_thresh = round(k * num_scored)

                pred_pos = pred_df.iloc[:num_above_thresh]
                
                
                precision = pred_pos.label_value.sum() / num_above_thresh if num_above_thresh > 0 else 0
                recall = pred_pos.label_value.sum() / num_positives if num_above_thresh > 0 else 0

                precisions.append(precision)
                recalls.append(recall)
                
            with sns.axes_style("darkgrid"): 
                if not only_recall:
                    sns.lineplot(x=k_values * 100, y=precisions, ax=ax, label='Precision@k')
                sns.lineplot(x=k_values * 100, y=recalls, ax=ax, label='Recall@k')
            

        ax.set_xlabel('Population percentage (k %)')
        ax.set_ylabel('Metric Value')
        ax.set_xlim(0, 100)
        ax.set_ylim(0, 1)
        # ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        # ax.set_title('Precision-Recall Curve')
        
        ax.legend(frameon=False)
        
        if title_string is None:
            ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        else:
            ax.set_title(title_string)
        return ax

    def plot_feature_importance(self, ax, n_top_features=20):
        """
        Plot the top most important individual features across model groups and train end times

        Args:
            ax: matplotlib Axes object to plot on
            n_top_features (int): the number of features to display
            
        Return:
            ax: the modified Axes object
        """

        feature_importance_scores = self.get_feature_importances(n_top_features=n_top_features)
        # keep only top n_top_features
        # feature_importance_scores.sort_values(by=['rank_abs'], ascending=True, inplace=True)
        # feature_importance_scores = feature_importance_scores.loc[feature_importance_scores['rank_abs'] <= n_top_features]
        
        if feature_importance_scores.empty:
            return ax


        if 'Algorithm does not support' in feature_importance_scores.feature.iloc[0]:
            # For models without featue importance score support
            return ax 
        
        # plot
        sns.barplot(
            data=feature_importance_scores,
            x='feature_importance', 
            y='feature',
            color='royalblue',
            ax=ax
        )
        ax.set_ylabel('Feature')
        ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        return ax

    def plot_feature_group_importance(self, ax, n_top_groups=20):
        """
        Plot the top most important feature groups as identified by the maximum importance of any feature in the group

        Args:
            ax: matplotlib Axes object to plot on

        Return: 
            ax: modified matplotlib Axes object
        """
        feature_group_importance = self.get_feature_group_importances()
        feature_group_importance.sort_values(by=['importance_aggregate'], ascending=False, inplace=True)
        feature_group_importance = feature_group_importance.reset_index(drop=True)
        feature_group_importance = feature_group_importance[feature_group_importance.index < n_top_groups] 
        sns.barplot(
            data=feature_group_importance,
            x='importance_aggregate', 
            y='feature_group',
            color='royalblue',
            ax=ax
        )
        ax.set_ylabel('Feature Group')
        ax.set_title(f'Model {self.model_id}, group: {self.model_group_id}')
        return ax

    def plot_calibration_curve(self, ax, nbins=20, min_score=0, max_score=1, matrix_uuid=None):

        """
        Plot the calibration curve of predicted probability scores across model groups

        Args:
            ax: matplotlib Axes object to plot on
            nbins (int): the number of bins to define the calibration curve
            min_score (optional, float between 0 and 1): zoom in on the plot, set x and y min axis value
            max_score (optional, float between 0 and 1): zoom in on the plot, set x and y max axis value
            matrix_uuid (optional): specify a matrix to get predictions for

        Return: 
            ax: modified matplotlib Axes object
        """
        # TODO can the bins be adjusted when zooming in on the plot?
        scores = self.get_predictions(matrix_uuid) # TODO what about null labels?
        scores = scores.dropna()
        # TODO do we want to restrict to label=1 only?
        cal_x, cal_y = calibration_curve(scores['label_value'], scores['score'], n_bins=nbins)
        sns.lineplot(
            x=cal_x,
            y=cal_y, 
            marker='o', 
            ax=ax
        )
        # plot perfectly calibrated line y = x for comparison
        probabilities = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        # filter out probabilities
        probabilities_plot = []
        for p in probabilities:
            probabilities_plot.append(p)
        sns.lineplot(
            data = pd.DataFrame(list(zip(probabilities, probabilities)), columns=['x', 'y']),
            x = 'x', 
            y = 'y', 
            color = 'gray',
            linestyle='dashed',
            ax=ax
        )
        ax.set_xlabel('Mean predicted probability')
        ax.set_ylabel('Fraction of data')
        ax.set_title(f'Model: {self.model_id}, Group: {self.model_group_id}')
        ax.set_xlim([min_score, max_score])
        ax.set_ylim([min_score, max_score])
        return ax
    

        




import pandas as pd
import numpy as np
import yaml
import os
import io
import psycopg2
import verboselogs, logging, coloredlogs
import matplotlib.pyplot as plt
import seaborn as sns
import graphviz

from sklearn.tree import DecisionTreeClassifier, plot_tree
from sklearn.model_selection import ParameterGrid
from sklearn.tree import export_text, export_graphviz

from triage.component.catwalk.storage import ProjectStorage

logger = verboselogs.VerboseLogger(__name__)


def _get_error_analysis_configuration():
    """
        Args:
            path (string): Path for the postmodeling_config.yaml file.

        Return:
            Dictionary of error analysis configuration
    """
    # TODO need to change it as in the rest of triage
    filename_yaml = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                                    "postmodeling_config.yaml") 
    with open(filename_yaml, 'r') as f:
        config = yaml.full_load(f)
    error_analysis_config = config['error_analysis']

    return error_analysis_config


def _get_random_seed(model_id, db_conn):
    """
        Args:
            model_id (int): The model id from to retrieve the random seed for.
            db_conn (sqlalchemy.engine.connect): Simple connection to the database.

        Returns:
            Integer of the random seed associated to this model_id.
    """
    q = """
        SELECT
            random_seed
        FROM triage_metadata.models
        WHERE model_id = {model_id}
    """.format(model_id=model_id)

    random_seed = pd.read_sql(q, db_conn)
    logger.debug("""obtained random seed associated to model_id: {model_id} for generating decision trees in error analysis""")

    return random_seed.random_seed.values[0]


def _fetch_scores_labels(model_id, db_conn):
    """
    Given a model id, it retrieves its scores and labels.

    Args:
        model_id (int): The model id from which scores and labels are going to 
        be retrieved from. 
        db_conn (sqlalchemy.engine.connect): Simple connection to the database.
    
    Returns:
        DataFrame with scores and labels for a specific model_id.
    """
    q = """
        SELECT
           model_id,
           entity_id,
           as_of_date,
           score,
           label_value,
           rank_abs_no_ties,
           matrix_uuid
        FROM test_results.predictions
        WHERE model_id = {model_id}
    """.format(model_id=model_id)
    
    df = pd.read_sql(q, db_conn)
    df.set_index(['entity_id', 'as_of_date'], inplace=True)

    return df


def _fetch_matrices(model_id, project_path, db_conn):
    """
    Joins the matrix of features, predictions and labels for a particular 
    model_id
    
    Args:
        model_id (int): Specific model id
        project_path (string): Where does the output of the model is stored
        (s3, file system)
        db_conn (sqlalchemy.engine.connect): Database engine connection

    Returns:
        A DataFrame with features, predictions, and labels
    """
    # getting predictions with label and matrix_uuid 
    logging.info(f"extracting predictions from model id: {model_id}")
    predictions = _fetch_scores_labels(model_id, db_conn)
    matrix_uuid = predictions.matrix_uuid.unique()[0]
    logging.info(f"matrix uuid: {matrix_uuid}")

    project_storage = ProjectStorage(project_path)
    matrix_storage_engine = project_storage.matrix_storage_engine()

    matrix_store = matrix_storage_engine.get_store(matrix_uuid=matrix_uuid)
    matrix = matrix_store.design_matrix

    # joining the predictions and labels for error analysis
    matrix_w_preds = predictions.join(matrix, how='left')
    logger.debug(f"predictions and labels joined for error analysis of model_id {model_id}")

    return matrix_w_preds


def _generate_error_labels(matrix, k):
    """
    Add columns to the original DataFrmae to define labels for the error model analysis.

    Args:
        matrix (DataFrame): DataFrame with the scores and labels.
        k (int): Number of resources to use
    
    Returns: 
        DataFrame with two new columns associated to the labels generated for each type of error: general error, 
        errors in positive label, errors in negative labels.
    """
    # extra df to create additional columns to the existing DF efficiently  
    # since the DF is "highly fragmentated" according to pandas.  
    prediction_col = ['0' for element in range(matrix.shape[0])]
    type_label_col = ['TP' for element in range(matrix.shape[0])]
    error_negative_label = ['0' for element in range(matrix.shape[0])]
    error_positive_label = ['0' for element in range(matrix.shape[0])]
    error_general_label = ['0' for element in range(matrix.shape[0])]

    # sort the scores desc
    sorted_scores = matrix.sort_values(by="rank_abs_no_ties")
    
    indexes = sorted_scores.index
    extra_df = pd.DataFrame({'prediction': prediction_col,
                            'type_label': type_label_col, 
                            'error_negative_label': error_negative_label,
                            'error_positive_label': error_positive_label,
                            'error_general_label': error_general_label 
                            }, index=indexes)
    
    data_df = pd.concat([sorted_scores, extra_df], axis=1)
    # add prediction column
    #sorted_scores['prediction'] = '0'
    data_df.loc[data_df.rank_abs_no_ties <= k, 'prediction'] = '1'
    # add type of label: TP, TN, FP, FN
    data_df['type_label'] = np.where(~(data_df.label_value) & 
                                     (data_df.prediction == '1'), 'FP', 'TP')
    data_df['type_label'] = np.where((data_df.label_value) & 
                                     (data_df.prediction == '0'), 'FN', 
                                     data_df.type_label)
    data_df['type_label'] = np.where(~(data_df.label_value) & 
                                     (data_df.prediction == '0'), 'TN', 
                                     data_df.type_label)
    
    # add three new columns with error analysis labels
    data_df['error_negative_label'] = np.where(data_df.type_label == 'FN', '1', '0')
    data_df['error_positive_label'] = np.where(data_df.type_label == 'FP', '1', '0')
    data_df['error_general_label'] = np.where((data_df.type_label == 'FP') | (data_df.type_label == 'FN'), '1', '0')
    logger.debug(f"error labels generated for k {k}")

    return data_df


def _error_analysis_model(model_id, matrix, grid, k, random_seed,
                          project_path, view_plots):
    """
    Generates the error analysis for a particular model_id by training Decision 
    Trees where the positive label is either: the false negatives, the 
    false positives, or both errors, false negatives and false positives. 

        Args: 
            model_id (int): Model id.
            matrix (DataFrame): Matrix of features.
            grid (ParamGrid): Grid of parameters for the Decision Trees
            project_path (string): Project path for storing output.
            view_plots (boolean): True if the plots are going to be plotted. 
            Default is False. In case of False, plots are stored as png files.

        Returns:
            List of dictionaries with results from all decision trees generated for 
            each configuration on each error type analysis
    """
    error_analysis_types = [element for element in list(matrix.columns) if element.startswith('error')]

    error_analysis_results = []
    for error_type in error_analysis_types:
        results = {'model_id': model_id, 'error_type': error_type, 'k': k}

        # first 5 columns of matrix aren't features
        predictions_cols = ['model_id', 'score', 'label_value', 'rank_abs_no_ties', 'matrix_uuid', 'prediction']
        no_features = predictions_cols + ['type_label'] + error_analysis_types

        if error_type == 'error_negative_label':
            X = matrix[matrix.type_label.isin(['TN', 'FN'])]\
                .drop(no_features, axis=1)
            y = matrix[matrix.type_label.isin(['TN', 'FN'])]\
                .filter([error_type], axis=1)
        elif error_type == 'error_positive_label':
            X = matrix[matrix.type_label.isin(['TP', 'FP'])]\
                .drop(no_features, axis=1)
            y = matrix[matrix.type_label.isin(['TP', 'FP'])]\
                .filter([error_type], axis=1)
        else:
            X = matrix.drop(no_features, axis=1)
            y = matrix.filter([error_type], axis=1)
        logger.debug(f"{error_type} k: {k} matrix size {X.shape}")
       
        for config in ParameterGrid(grid):
            error_model = None
            config_params = {}
            max_depth = config['max_depth']
            dt = DecisionTreeClassifier(max_depth=max_depth, random_state=random_seed)
            error_model = dt.fit(X, y)

            feature_importances_ = error_model.feature_importances_
            importances_idx = np.argsort(-feature_importances_)
            # retrieving all features with importance > 0
            top_n = [i for i, element in enumerate(importances_idx) if feature_importances_[element] == 0][0]
            logger.debug(f"feature importance indices {importances_idx[:top_n]}")
            feature_ = error_model.feature_names_in_
            feature_names_importance_sorted = feature_[importances_idx[:top_n]]
            
            config_params['tree'] = error_model
            config_params['max_depth'] = max_depth
            config_params['feature_importance'] = feature_importances_[
                                                    importances_idx[:top_n]]
            config_params['feature_names'] = feature_names_importance_sorted
            config_params['feature_names_graphviz'] = list(X.columns.values)
            config_params['tree_text'] = export_text(error_model,
                                                     feature_names=list(feature_),
                                                     show_weights=True)
            
            _plot_feature_importance(config_params['feature_importance'], 
                                     config_params['feature_names'],
                                     error_type, 
                                     project_path,
                                     k,
                                     max_depth, 
                                     model_id, 
                                     view_plots)
        
            _generate_tree_graphviz(error_model,
                                    error_type,
                                    list(X.columns.values), 
                                    k, 
                                    max_depth, 
                                    project_path,
                                    model_id,
                                    view_plots)

            error_analysis_results.append({**results, **config_params})
            logger.debug(f"error analysis generated for model_id {model_id}, with k {k}")

    return error_analysis_results


def generate_error_analysis(model_id, db_conn, project_path):
    """
    Main function that runs the error analysis on a particular model_id.

        Args: 
            model_group_ids (list): List of model groups ids 
            db_conn (SQLAlchemy simple connection): Database engine connection
            view_plots (boolean): If plots are going to be displayed (True) or saved as png files (False). 
        
        Returns:
            List of list of dictionaries
    """
    error_analysis_config = _get_error_analysis_configuration()
    #project_path = error_analysis_config['project_path']
    k_set = error_analysis_config['k']
    grid = error_analysis_config['model_params']
    view_plots = error_analysis_config['view_plots']
    
    matrix_data = _fetch_matrices(model_id, project_path, db_conn)
    error_analysis_results = []
    for k in k_set:
        new_matrix = _generate_error_labels(matrix_data, k)
        random_seed = _get_random_seed(model_id, db_conn)
        error_analysis_result = _error_analysis_model(model_id,
                                                      new_matrix,
                                                      grid,
                                                      k,
                                                      random_seed,
                                                      project_path,
                                                      view_plots)
        error_analysis_results.append(error_analysis_result)
    #TODO where and how to save the outputs of the errror analysis (list of dicts)
    logger.debug(f"error analysis finished for model id {model_id}")

    return error_analysis_results


def _plot_feature_importance(importances, features, error_type, project_path, 
                                k, max_depth, model_id, view_plots):
    """
    Plots a seborn barplot with the most important features associated to the 
    error in the label analyzed.

        Args:
            importances (list): List with the importance value of the top n 
                                important features
            features (list): List with the name of the n most important features
            error_label (string): Type of error analysis 
            project_path (string): Project path for store output
            k (int): Size of list
            max_depht (int): Max depth of tree.
            model_id (int): Model id associated to this error analysis.
            view_plots (boolean): True if the plot is going to be displayed, False 
             if it is going to be save.
    """
    error_label = _get_error_label(error_type)
    
    plt.figure()
    df_viz = pd.DataFrame({'importance': importances, 'feature': features})
    df_viz.sort_values(by="importance", ascending=False, inplace=True)
    if df_viz.shape[0] > 10: 
        df_viz_ = df_viz.head(10)
    else: 
        df_viz_ = df_viz 

    sns.barplot(data=df_viz_, 
                x="importance", y="feature", color="grey")
    plt.title("Top feature importance in " + error_label + " analysis.")

    if view_plots: 
        plt.show()
    else: 
        # save plot 
        feature_importance_filename = "_".join(['feature_importance',
                                                error_type,
                                                str(k), 
                                                str(max_depth)])
        feature_importance_filename += ".png"
        # stream of plot
        img_data = io.BytesIO()
        plt.savefig(img_data, format="png", dpi=300, bbox_inches="tight")
        img_data.seek(0)

        project_storage = ProjectStorage(project_path)
        storage = project_storage.get_store(["error_analysis/feature_importance/" +
                                            str(model_id)], 
                                            feature_importance_filename)
        storage.write(img_data.read())
        plt.close()


def _plot_feature_importance_local(importances, features, error_label):
    """
    Plots a seborn barplot with the most important features associated to the 
    error in the label analyzed.

        Args:
            importances (list): List with the importance value of the top n 
                                important features
            features (list): List with the name of the n most important features
            error_label (string): Type of error analysis 
            project_path (string): Project path for store output 
    """
    plt.clf()
    df_viz = pd.DataFrame({'importance': importances, 'feature': features})
    df_viz.sort_values(by="importance", ascending=False, inplace=True)
    if df_viz.shape[0] > 10: 
        df_viz_ = df_viz.head(10)
    else: 
        df_viz_ = df_viz 
    
    a = sns.barplot(data=df_viz_,
                        x="importance", y="feature", color="grey")
    a.set_title("Top feature importance in " + error_label + " analysis.")
    plt.show()


def _generate_tree_graphviz(error_model, error_type, feature_names, k, 
                            max_depth, project_path, model_id, view_plots):
    """
    Generates a png with the tree rules generated for the error analysis by 
    using graphviz. The output is stored in the storage associated to the project.

        Args: 
            error_model (DecisionTreeClasifier): The Decission Tree classifier 
             trained error_type (string): Type or error analysis made 
            feature_names (List): List of all the feature names in the train 
             matrix
            k (int): Size of the list
            max_depth (int): Max depth of the decision tree classifier trained
            project_path (string): General project path
            model_id (int): Model id associated to the error analysis
            view_plots (boolean): True if the plot is going to be displayed, False 
             if it is going to be save.
    """
    dot_path = "_".join(["tree", error_type, str(k), str(max_depth)])
    dot_path += ".png"
    tree_viz = export_graphviz(error_model,
                               out_file=None,
                               feature_names=feature_names,
                               filled=True,
                               special_characters=True)
    
    if view_plots: 
        plot_tree(error_model, filled=True)
    else:
        graph = graphviz.Source(tree_viz) 
        img_data = io.BytesIO()
        img_data.write(graph.pipe("png"))
        img_data.seek(0)

        project_storage = ProjectStorage(project_path)
        storage = project_storage.get_store(["error_analysis/tree/" + 
                                             str(model_id)], dot_path)
        storage.write(img_data.read())


def _output_config_text(element, error_label):    
    """
    First paragraph of text associated to the output of a particular analysis.

        Args:
            element (dict): Dictionary with outputs of the error analysis.
            error_label (string): Type of error analysis that is being generated.

        Returns:
            dt_text (string): Text associated to the output of the error 
            analysis generated.
    """
    config_text = f"""
        
    Model id: {element['model_id']}
    Error analysis type: {error_label}
    Size of the list: {element['k']}
    Decision Tree with max_depth of, {element['max_depth']}

    Top feature importance associated with error in label type {error_label}
    """
    return config_text


def _output_dt_text(error_label):
    """
    Text associated to the output of the decission trees from a particular anlaysis. 

        Args: 
            error_label (string): Type of error analysis that is being generated.

        Returns: 
            dt_text (string): Text associated to the output of the error analysis 
            generated. 
    """
    dt_text = f"""
    Rules made with the top features associated with {error_label}
    """

    return dt_text
  

def _get_error_label(error_type):
    """
    Given the type of error analysis, returns a descriptive name. 

        Args: 
            error_type (string): Type of error analysis. 
        
        Returns:
            error_label (string): Human readable output related to the type 
            of error analysis.  
    """
    if error_type == 'error_negative_label':
        error_label = 'Mistakes assigning Negative labels (FN)'
    elif error_type == 'error_positive_label':
        error_label = 'Mistakes assigning Positive labels (FP)'
    else:
        error_label = 'Mistakes assigning Positive and Negative (FP & FN)'

    return error_label


def output_all_analysis(error_analysis_results):
    """
    Prints as text the output of all the error analysis made.

        Args: 
            error_anlaysis_resulst (list): List of dictionaries with the output 
            of every single error anlaysis made.
    """
    print("Error Analysis")
    print("--------------")

    for analysis in error_analysis_results:
        for element in analysis:
            error_label = _get_error_label(element['error_type'])
            config_text = _output_config_text(element, error_label)
            print(config_text)
            
            _plot_feature_importance_local(element['feature_importance'],
                                           element['feature_names'],
                                           error_label)

            dt_text = _output_dt_text(error_label)
            print(dt_text)
            plot_tree(element['tree'], filled=True, feature_names=
                      element['feature_names_graphviz'])
            plt.show()
            #print(element['tree_text'])
            print("             ######            ")
        
        print("*******************************************")


def output_specific_configuration(error_analysis_resutls, 
                                  k=100, 
                                  max_depth=10,
                                  error_type="error_negative_label"):
    """
    Prints as text the output of a specific configuration of error analysis made.

        Args: 
            error_analysis_results (list): List of dictionaries with all the 
            output of the error analysis. 
            k (int): A specific size of the list. Default `100`
            max_depth (int): A specific depth of decision trees. Default `10`
            error_type (string): Type of error analysis. Default `error_negative_label` 
    """
    for analysis in error_analysis_resutls:
        for element in analysis:
            if ((element['error_type'] == error_type) &
                (element['k'] == k) & 
                (element['max_depth'] == max_depth)):
                error_label = _get_error_label(error_type)
                config_text = _output_config_text(element, error_label)
                print(config_text)
                
                _plot_feature_importance_local(element['feature_importance'],
                                               element['feature_names'],
                                               error_label)
                
                dt_text = _output_dt_text(error_label)
                print(dt_text)
                plot_tree(element['tree'], filled=True,
                          feature_names=element['feature_names_graphviz'])
                plt.show()
                #print(element['tree_text'])
                print("             ######            ")


def output_specific_error_analysis(error_analysis_results, 
                                   error_type='error_negative_label'):
    """
    Prints all the outputs associated to a specific type of error analysis made: 
    negative label, positive label, or both. 

        Args: 
            error_analysis_results (list): List of dictionaries with the outputs
            of the error analysis. 
            error_type (string): Type of error analysis.
    """
    for analysis in error_analysis_results:
        for element in analysis:
            if element['error_type'] == error_type:
                error_label = _get_error_label(error_type)
                config_text = _output_config_text(element, error_label)
                print(config_text)
                
                _plot_feature_importance_local(element['feature_importance'], 
                                               element['feature_names'], 
                                               error_label)

                dt_text = _output_dt_text(error_label)
                print(dt_text)
                plot_tree(element['tree'], filled=True,
                          feature_names=element['feature_names_graphviz'])
                plt.show()
                #print(element['tree_text'])
                print("             ######            ")
        
        print("*******************************************")


if __name__ == "__main__":
    model_id = 1417
    db_conn = psycopg2.connect(service='acdhs_housing')
    # its given by the repor generator 
    project_path = 's3://dsapp-social-services-migrated/acdhs_housing/triage_experiments/' 

    generate_error_analysis(model_id, db_conn, project_path)



"""merge b097e47ba829 with cdd0dc9d9870

Revision ID: 079a74c15e8b
Revises: b097e47ba829, cdd0dc9d9870
Create Date: 2021-05-30 20:49:19.039280

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '079a74c15e8b'
down_revision = ('b097e47ba829', 'cdd0dc9d9870')
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import csv
import datetime
import hashlib
import numpy as np
import os
import pandas as pd
import json

import verboselogs, logging

logger = verboselogs.VerboseLogger(__name__)

import random
from itertools import chain
from functools import partial

import postgres_copy
import sqlalchemy
from retrying import retry
from sqlalchemy.orm import sessionmaker
from ohio import PipeTextIO

from triage.component.results_schema import (
    Experiment,
    Matrix,
    Model,
    ExperimentMatrix,
    ExperimentModel,
    TriageRun,
)


def filename_friendly_hash(inputs):
    def dt_handler(x):
        if isinstance(x, datetime.datetime) or isinstance(x, datetime.date):
            return x.isoformat()
        raise TypeError("Unknown type")

    return hashlib.md5(
        json.dumps(inputs, default=dt_handler, sort_keys=True).encode("utf-8")
    ).hexdigest()


def get_subset_table_name(subset_config):
    return "subset_{}_{}".format(
        subset_config.get("name", "default"),
        filename_friendly_hash(subset_config),
    )


def retry_if_db_error(exception):
    return isinstance(exception, sqlalchemy.exc.OperationalError)


DEFAULT_RETRY_KWARGS = {
    "retry_on_exception": retry_if_db_error,
    "wait_exponential_multiplier": 1000,  # wait 2^x*1000ms between each retry
    "stop_max_attempt_number": 14,
    # with this configuration, last wait will be ~2 hours
    # for a total of ~4.5 hours waiting
}


db_retry = retry(**DEFAULT_RETRY_KWARGS)


@db_retry
def save_experiment_and_get_hash(config, db_engine):
    experiment_hash = filename_friendly_hash(config)
    session = sessionmaker(bind=db_engine)()
    session.merge(Experiment(experiment_hash=experiment_hash, config=config))
    session.commit()
    session.close()
    return experiment_hash


@db_retry
def associate_matrices_with_experiment(experiment_hash, matrix_uuids, db_engine):
    session = sessionmaker(bind=db_engine)()
    for matrix_uuid in matrix_uuids:
        session.merge(
            ExperimentMatrix(experiment_hash=experiment_hash, matrix_uuid=matrix_uuid)
        )
    session.commit()
    session.close()
    logger.spam("Associated matrices with experiment in database")


@db_retry
def associate_models_with_experiment(experiment_hash, model_hashes, db_engine):
    session = sessionmaker(bind=db_engine)()
    for model_hash in model_hashes:
        session.merge(
            ExperimentModel(experiment_hash=experiment_hash, model_hash=model_hash)
        )
    session.commit()
    session.close()
    logger.spam("Associated models with experiment in database")


@db_retry
def missing_matrix_uuids(experiment_hash, db_engine):
    """Compare the contents of the experiment_matrices table with that of the
    matrices table to produce a list of matrix_uuids that the experiment wants
    but are not available.
    """
    query = f"""
        select experiment_matrices.matrix_uuid
        from {ExperimentMatrix.__table__.fullname} experiment_matrices
        left join {Matrix.__table__.fullname} matrices
        on (experiment_matrices.matrix_uuid = matrices.matrix_uuid)
        where experiment_hash = %s
        and matrices.matrix_uuid is null
    """
    return [row[0] for row in db_engine.execute(query, experiment_hash)]


@db_retry
def missing_model_hashes(experiment_hash, db_engine):
    """Compare the contents of the experiment_models table with that of the
    models table to produce a list of model hashes the experiment wants
    but are not available.
    """
    query = f"""
        select experiment_models.model_hash
        from {ExperimentModel.__table__.fullname} experiment_models
        left join {Model.__table__.fullname} models
        on (experiment_models.model_hash = models.model_hash)
        where experiment_hash = %s
        and models.model_hash is null
    """
    return [row[0] for row in db_engine.execute(query, experiment_hash)]


class Batch:
    # modified from
    # http://codereview.stackexchange.com/questions/118883/split-up-an-iterable-into-batches
    def __init__(self, iterable, limit=None):
        self.iterator = iter(iterable)
        self.limit = limit
        try:
            self.current = next(self.iterator)
        except StopIteration:
            self.on_going = False
        else:
            self.on_going = True

    def group(self):
        yield self.current
        # start enumerate at 1 because we already yielded the last saved item
        for num, item in enumerate(self.iterator, 1):
            self.current = item
            if num == self.limit:
                break
            yield item
        else:
            self.on_going = False

    def __iter__(self):
        while self.on_going:
            yield self.group()


AVAILABLE_TIEBREAKERS = {"random", "best", "worst"}


def sort_predictions_and_labels(
    predictions_proba, labels, df_index, tiebreaker="random", sort_seed=None
):
    """Sort predictions and labels with a configured tiebreaking rule

    Args:
        predictions_proba (np.array) The predicted scores
        labels (np.array) The numeric labels (1/0, not True/False)
        df_index (pd.MultiIndex) Index (generally entity_id, as_of_date tuples) to be sorted with the labels/scores
        tiebreaker (string) The tiebreaking method ('best', 'worst', 'random')
        sort_seed (signed int) The sort seed. Needed if 'random' tiebreaking is picked.

    Returns:
        (tuple) (predictions_proba, labels, df_index), sorted
    """
    if len(labels) == 0:
        logger.notice("No labels present, skipping predictions sorting .")
        return (predictions_proba, labels, df_index)

    df = pd.DataFrame(predictions_proba, columns=["score"])
    df["label_value"] = labels
    df.set_index(df_index, inplace=True)

    if tiebreaker == "random":
        if not sort_seed:
            raise ValueError("If random tiebreaker is used, a sort seed must be given")
        random.seed(sort_seed)
        np.random.seed(sort_seed)
        df["random"] = np.random.rand(len(df))
        df.sort_values(by=["score", "random"], inplace=True, ascending=[False, False])
        df.drop("random", axis=1)
    elif tiebreaker == "worst":
        df.sort_values(
            by=["score", "label_value"],
            inplace=True,
            ascending=[False, True],
            na_position="first",
        )
    elif tiebreaker == "best":
        df.sort_values(
            by=["score", "label_value"],
            inplace=True,
            ascending=[False, False],
            na_position="last",
        )
    else:
        raise ValueError(f"Unknown tiebreaker: {tiebreaker}")

    return [df["score"].to_numpy(), df["label_value"].to_numpy(), df.index]


@db_retry
def retrieve_model_id_from_hash(db_engine, model_hash):
    """Retrieves a model id from the database that matches the given hash

    Args:
        db_engine (sqlalchemy.engine) A database engine
        model_hash (str) The model hash to lookup

    Returns: (int) The model id (if found in DB), None (if not)
    """
    session = sessionmaker(bind=db_engine)()
    try:
        saved = session.query(Model).filter_by(model_hash=model_hash).one_or_none()
        return saved.model_id if saved else None
    finally:
        session.close()


@db_retry
def retrieve_model_hash_from_id(db_engine, model_id):
    """Retrieves the model hash associated with a given model id

    Args:
        model_id (int) The id of a given model in the database

    Returns: (str) the stored hash of the model
    """
    session = sessionmaker(bind=db_engine)()
    try:
        return session.query(Model).get(model_id).model_hash
    finally:
        session.close()


@db_retry
def retrieve_existing_model_random_seeds(
    db_engine,
    model_group_id,
    train_end_time,
    train_matrix_uuid,
    training_label_timespan,
    experiment_random_seed,
):
    """Retrieve existing model random seeds matching the model parameters and
    experiment-level random seed to allow for reusing seeds before creating a
    new one.
    """
    query = f"""
        select models.random_seed
        from {ExperimentModel.__table__.fullname} experiment_models
        join {Model.__table__.fullname} models
        on (experiment_models.model_hash = models.model_hash)
        join {TriageRun.__table__.fullname} triage_runs
        on (
            experiment_models.experiment_hash = triage_runs.run_hash
            --makes sure that the model has been created before, and not being created in this run
            and models.built_in_triage_run = triage_runs.id
        )
        where models.model_group_id = {model_group_id}
        and models.train_end_time = '{train_end_time}'
        and models.train_matrix_uuid = '{train_matrix_uuid}'
        and models.training_label_timespan = '{training_label_timespan}'
        and triage_runs.random_seed = {experiment_random_seed}
        order by models.run_time DESC, random()
    """
    logging.debug(f"Query that will retrieve random seeds for the model: {query}")
    return [
        row[0]
        for row in db_engine.execute(query)
    ]


@db_retry
def retrieve_experiment_seed_from_run_id(db_engine, run_id):
    """Retrieves the random seed associated with a given experiment run

    Args:
        run_id (int) The id of a given experiment run in the database

    Returns: (int) the stored random seed from the experiment
    """
    session = sessionmaker(bind=db_engine)()
    try:
        return session.query(TriageRun).get(run_id).random_seed
    finally:
        session.close()


def _write_csv(file_like, db_objects, type_of_object):
    writer = csv.writer(file_like, quoting=csv.QUOTE_MINIMAL, lineterminator="\n")
    for db_object in db_objects:
        if type(db_object) != type_of_object:
            raise TypeError(
                "Cannot copy collection of objects to db as they are not all "
                f"of the same type. First object was {type_of_object} "
                f"and later encountered a {type(db_object)}"
            )
        writer.writerow(
            [getattr(db_object, col.name) for col in db_object.__table__.columns]
        )


@db_retry
def save_db_objects(db_engine, db_objects):
    """Saves a collection of SQLAlchemy model objects to the database using a COPY command

    Args:
        db_engine (sqlalchemy.engine)
        db_objects (iterable) SQLAlchemy model objects, corresponding to a valid table
    """
    db_objects = iter(db_objects)
    first_object = next(db_objects)
    type_of_object = type(first_object)
    columns = [col.name for col in first_object.__table__.columns]

    with PipeTextIO(
        partial(
            _write_csv,
            db_objects=chain((first_object,), db_objects),
            type_of_object=type_of_object,
        )
    ) as pipe:
        postgres_copy.copy_from(
            source=pipe,
            dest=type_of_object,
            engine_or_conn=db_engine,
            columns=columns,
            format="csv",
        )



"""add run_type

Revision ID: 5dd2ba8222b1
Revises: 079a74c15e8b
Create Date: 2021-07-22 23:53:04.043651

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '5dd2ba8222b1'
down_revision = '079a74c15e8b'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('experiment_runs', sa.Column('run_type', sa.Text(), nullable=True), schema='triage_metadata')
    op.execute("UPDATE triage_metadata.experiment_runs SET run_type='experiment' WHERE run_type IS NULL")

    op.alter_column('experiment_runs', 'experiment_hash', nullable=True, new_column_name='run_hash', schema='triage_metadata')
    op.drop_constraint('experiment_runs_experiment_hash_fkey', 'experiment_runs', type_='foreignkey', schema='triage_metadata')

    op.execute("ALTER TABLE triage_metadata.experiment_runs RENAME TO triage_runs")

    op.create_table('retrain',
        sa.Column('retrain_hash', sa.Text(), nullable=False),
        sa.Column('config', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('prediction_date', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('retrain_hash'),
        schema='triage_metadata',
    )

    op.alter_column('models', 'built_in_experiment_run', nullable=False, new_column_name='built_in_triage_run', schema='triage_metadata')
    op.execute("CREATE TABLE triage_metadata.deprecated_models_built_by_experiment AS SELECT model_id, model_hash, built_by_experiment FROM triage_metadata.models")
    op.drop_column('models', 'built_by_experiment', schema='triage_metadata')

    op.create_table('retrain_models',
        sa.Column('retrain_hash', sa.String(), nullable=False),
        sa.Column('model_hash', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(['retrain_hash'], ['triage_metadata.retrain.retrain_hash'], ),
        sa.PrimaryKeyConstraint('retrain_hash', 'model_hash'),
        schema='triage_metadata'
    )


def downgrade():
    op.execute("ALTER TABLE triage_metadata.triage_runs RENAME TO experiment_runs")
    op.drop_column('experiment_runs', 'run_type', schema='triage_metadata')
    op.alter_column('experiment_runs', 'run_hash', nullable=True, new_column_name='experiment_hash', schema='triage_metadata')
    op.create_foreign_key('experiment_runs_experiment_hash_fkey', 'experiment_runs', 'experiments', ['experiment_hash'], ['experiment_hash'], source_schema='triage_metadata', referent_schema='triage_metadata')
    op.drop_table('retrain_models', schema='triage_metadata')
    op.drop_table('retrain', schema='triage_metadata')
    op.add_column('models', sa.Column('built_by_experiment', sa.Text(), nullable=True), schema='triage_metadata')
    op.alter_column('models', 'built_in_triage_run', nullable=False, new_column_name='built_in_experiment_run', schema='triage_metadata')




"""rename production schema and list_predcitons to triage_predcition and predictions 

Revision ID: cdd0dc9d9870
Revises: 670289044eb2
Create Date: 2021-04-13 00:53:56.098572

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'cdd0dc9d9870'
down_revision = '670289044eb2'
branch_labels = None
depends_on = None


def upgrade():
    op.execute("CREATE SCHEMA IF NOT EXISTS triage_production")
    op.execute("ALTER TABLE production.list_predictions SET SCHEMA triage_production;")
    op.execute("ALTER TABLE production.prediction_metadata SET SCHEMA triage_production")
    op.execute("ALTER TABLE triage_production.list_predictions RENAME TO predictions")
 

def downgrade():
    op.execute("ALTER TABLE triage_production.predictions SET SCHEMA production;")
    op.execute("ALTER TABLE triage_production.prediction_metadata SET SCHEMA production")
    op.execute("ALTER TABLE production.predictions RENAME TO list_predictions")
    op.execute("DROP SCHEMA IF EXISTS triage_production")



""" This is a temporaty script for writing functions that analyse the top-k lists,
    while others are working on other components. Eventually these files will be merged
"""

import logging
import itertools
import pandas as pd

# remove this
import psycopg2

from scipy.stats import spearmanr

from triage.component.catwalk.storage import ProjectStorage


# TODO: Modify how the threshold is specified
def get_highest_risk_entities(db_engine, model_id, threshold, include_all_ties=False):
    """ Fetch the entities with the highest risk for a particular model
        
        args:
            db_engine (sqlalchemy.engine)   : Database connection engine
            model_id (int)  : Model ID we are interested in
            threshold (Union[int, float]): value of the threshold (k)
            include_all_ties (bool): Whether to include all entities with the same score in the list (can include more than k elements) 
                                    Defaults to False where ties are broken randomly

        return:
            a pd.DataFrame object that contains the entity_id, as_of_date, the hash of the text matrix, model score, 
            and all rank columns 
            
    """
    
    # Determining which rank column to filter by
    ties_suffix = 'no_ties'
    col = 'rank_pct'
    
    if include_all_ties:
        ties_suffix='with_ties'

    if isinstance(threshold, int):
        col = 'rank_abs'      

    col_name = f'{col}_{ties_suffix}'

    q = f"""
        select 
            entity_id, 
            as_of_date,
            score,
            label_value,
            rank_abs_no_ties,
            rank_abs_with_ties,
            rank_pct_no_ties,
            rank_pct_with_ties,
            matrix_uuid
        from test_results.predictions
        where model_id={model_id}
        and {col_name} <= {threshold}
    """

    top_k = pd.read_sql(q, db_engine)

    return top_k
   

def pairwise_comparison_model_lists(db_engine, models, k, include_all_tied_entities=False):
    """ Given two lists compare their similarities

        args:
            db_engine (sqlalchemy.engine): Database connection engine 
            models (List[int]): A list of model_ids that we are interested in comparing
            k (int): The list size
            include_all_tied_entities (bool): Whether to include all entities with the same score in the list 
                        If True, can include more than k elements in the list.  
                        Defaults to False where ties are broken randomly
 
        return:
            pairwise list comparison metrics (Dict[Dict]): Dictionay of dictionaries 
                                that contains 'overlap', 'jaccard_similarity', 'rank_corr' for each model_id pair
    """
    
    logging.info('Fetching the top-k lists for all models')
    risk_lists = dict()
    for model_id in models:
        risk_lists[model_id] = get_highest_risk_entities(db_engine, model_id, k, include_all_tied_entities)

    pairs = list(itertools.combinations(risk_lists.keys(), 2))

    logging.info(f'You provided {len(models)} models. Performing {len(pairs)} comparisons')

    # Storing all results as a dictionary of dictionaries
    # Key model_id pair-- Tuple[(int, int)], Value -- Dict[jaccard, overlap, rank_corr]
    results = dict()

    for model_pair in pairs:
        logging.info(f'Comparing {model_pair[0]} and {model_pair[1]}')

        # Dictionary that stores the comparison metrics 
        d = dict()
    
        df1 = risk_lists[model_pair[0]]
        df2 = risk_lists[model_pair[1]]

        if df1.as_of_date.at[0] != df2.as_of_date.at[0]:
            logging.warning('You are comparing two lists generated on different as_of_dates!')

        # calculating jaccard similarity and overlap
        entities_1 = set(df1.entity_id)
        entities_2 = set(df2.entity_id)

        inter = entities_1.intersection(entities_2)
        un = entities_1.union(entities_2)    
        d['jaccard'] = len(inter)/len(un)

        # If the list sizes are not equal, using the smallest list size to calculate simple overlap
        d['overlap'] = len(inter)/ min(len(entities_1), len(entities_2))

        # calculating rank correlation
        df1.sort_values('score', ascending=False, inplace=True)
        df2.sort_values('score', ascending=False, inplace=True)

        # only returning the corr coefficient, not the p-value
        d['rank_corr'] = spearmanr(df1.entity_id.iloc[:, 0], df2.entity_id.iloc[:, 1])[0]

        return results


def _fetch_relevant_matrix_hashes(db_engine, model_id):
    """"""

    q = f"""
        select 
            train.matrix_uuid as train_matrix_uuid,
            test.matrix_uuid as test_matrix_uuid
        from train_results.prediction_metadata train join test_results.prediction_metadata test
        using(model_id) 
        where model_id={model_id};
    """

    matrices = pd.read_sql(q, db_engine).to_dict(orient='records')

    return matrices


# Currently this function only calculates the mean ratio
# NOTE This code was incorported to the ModelAnalyzer
def get_crosstabs_postive_vs_negative(
    db_engine, 
    model_id, 
    project_path, 
    thresholds,  
    return_df=True, 
    matrix_uuid=None, 
    table_name='crosstabs'
):
    """ For a given model_id, generate crosstabs for the top_k vs the rest
    
        args:
            model_id (int): The model_id we are intetereted in
            project_path (str): Path where the experiment artifacts (models and matrices) are stored
            thresholds (Dict{str: Union[float, int}]): A dictionary that maps threhold type to the threshold
                                                    The threshold type can be one of the rank columns in the test_results.predictions_table
            return_df (bool, optional): Whether to return the constructed df or just to store in the database
                                        Defaults to False (only writing to the db)
            table_name (str, optional): Table name to use in the db's `test_results` schema. Defaults to crosstabs
            matrix_uuid (str, optional): If we want to run crosstabs for a different matrix than the validation matrix from the experiment
    """

    # Table structure
    # model_id, matrix_uuid, 
    # threshold_type, threshold, feature, metric, value
    

    if matrix_uuid is None:
        matrix_uuid = _fetch_relevant_matrix_hashes(db_engine, model_id)[0]['test_matrix_uuid']

    logging.info('Fetching predictions for the model')
    
    # NOTE/TODO If we use a Model object here, we can avoid these repeated db calls
    q = f"""
        select 
            entity_id, 
            as_of_date,
            score,
            label_value,
            rank_abs_no_ties,
            rank_abs_with_ties,
            rank_pct_no_ties,
            rank_pct_with_ties,
            matrix_uuid
        from test_results.predictions
        where model_id={model_id} and matrix_uuid = '{matrix_uuid}'
    """

    predictions = pd.read_sql(q, db_engine)
    predictions.set_index(['entity_id', 'as_of_date'], inplace=True)

    if predictions.empty:
        logging.error(f'No predictions found for {model_id} and matrix {matrix_uuid}. Exiting!')
        raise ValueError(f'No predictions found {model_id} and matrix {matrix_uuid}')


    # initializing the storage engines
    project_storage = ProjectStorage(project_path)
    matrix_storage_engine = project_storage.matrix_storage_engine()

    matrix_store = matrix_storage_engine.get_store(matrix_uuid=matrix_uuid)
    matrix = matrix_store.design_matrix
    labels = matrix_store.labels
    features = matrix.columns

    # joining the predictions to the model
    matrix = predictions.join(matrix, how='left')

    for threshold_name, threshold in thresholds.items():
        logging.debug('')
        msk = matrix[threshold_name] <= threshold
        postive_preds = matrix[msk]
        negative_preds = matrix[~msk]

        # TODO: Take a list of metrics to calculate and iterate (as it's done in crosstabs.py)

        # Calculates the mean ratio for each feature and produces a series indexed by the feature n,e
        mean_ratios = (postive_preds[features].mean() / negative_preds[features].mean()).reset_index()
        mean_ratios['metric'] = 'mean_ratio_pos_over_neg'
        
        non_zero_rows_count_pos_pred = (postive_preds[features] > 0).sum().reset_index()
        non_zero_rows_count_pos_pred['metric'] = 'non_zero_rows_pos_pred_count'

        non_zero_rows_frac_pos_pred = (postive_preds[features] > 0).mean().reset_index()
        non_zero_rows_frac_pos_pred['metric'] = 'non_zero_rows_pos_pred_pct'
        
        non_zero_rows_count_neg_pred = (negative_preds[features] > 0).sum().reset_index()
        non_zero_rows_count_neg_pred['metric'] = 'non_zero_rows_neg_pred_count'

        non_zero_rows_frac_neg_pred = (negative_preds[features] > 0).mean().reset_index()
        non_zero_rows_frac_neg_pred['metric'] = 'non_zero_rows_pos_pred_pct'

        crosstabs_df = pd.concat([
            mean_ratios, 
            non_zero_rows_count_pos_pred, 
            non_zero_rows_count_neg_pred, 
            non_zero_rows_frac_pos_pred,
            non_zero_rows_frac_neg_pred
        ])

        crosstabs_df.rename(columns={'index': 'feature', 0: 'value'}, inplace=True)
        crosstabs_df['model_id'] = model_id
        crosstabs_df['matrix_uuid'] = matrix_uuid
        crosstabs_df['threshold_type'] = threshold_name
        crosstabs_df['threshold'] = threshold

        crosstabs_df.to_csv('temp_crosstabs.csv', index=False)

        if return_df:
            return crosstabs_df
        


def _get_descriptives(db_engine, model_id, columns_of_interest, feature_groups):
    """ Given a list of entities, generated descriptives

        args:
            entities (pd.DataFrame): The Dataframe containing information of the entities
            columns_of_interest (List[str]): The list of column names we are interested in describing
    """

    # NOTE: As the first pass, we'll calculate descriptives of 


if __name__ == '__main__':

    conn = psycopg2.connect(service='acdhs_housing')

    df = get_crosstabs_postive_vs_negative(
        conn,
        model_id=1533,
        project_path='s3://dsapp-social-services-migrated/acdhs_housing/triage_experiments/',
        thresholds={'rank_abs_no_ties': 100},
        table_name='crosstabs_test_kasun'
    )

    print(df.sample(n=10))